import matplotlib.pyplot as plt
import numpy as np

datadir='../output/HM_IA/'
fig, ((ax1, ax2)) = plt.subplots(1, 2,figsize=(10,4))

def plot_pow_spec(ax,pow_spec,iz,colour,linestyle):
    # Read in the power spectra files
    kvals = np.loadtxt(datadir+pow_spec+'/k_h.txt')
    pkvals = np.loadtxt(datadir+pow_spec+'/p_k.txt')
    zvals = np.loadtxt(datadir+pow_spec+'/z.txt')
    ax.plot(kvals, np.abs(pkvals[iz,:]),label=pow_spec,color=colour,linestyle=linestyle)
#    ax.plot(kvals, kvals*kvals*pkvals[iz,:],label=pow_spec,color=colour,linestyle=linestyle)
    print(np.max(pkvals[iz,:]),np.min(pkvals[iz,:]),zvals[iz])

def plot_ratio_pow_spec(ax,pow_spec,pkvals_base,iz,colour,linestyle):
    # Read in the power spectra files
    kvals = np.loadtxt(datadir+pow_spec+'/k_h.txt')
    pkvals = np.loadtxt(datadir+pow_spec+'/p_k.txt')
    zvals = np.loadtxt(datadir+pow_spec+'/z.txt')
    ax.plot(kvals, pkvals[iz,:]/pkvals_base[iz,:],label=pow_spec,color=colour,linestyle=linestyle)
    print(np.max(pkvals[iz,:]),zvals[iz])

# Choose redshift of interest
iz=2  #z=0.75
plot_pow_spec(ax1,'matter_power_nl',iz,'black','solid')
plot_pow_spec(ax1, 'galaxy_power_red',iz,'red','solid')
plot_pow_spec(ax1, 'galaxy_power_blue',iz,'blue','solid')
#plot_pow_spec(ax1,'matter_intrinsic_power',iz,'grey','solid')
#plot_pow_spec(ax1,'matter_intrinsic_power_blue',iz,'blue','solid')
#plot_pow_spec(ax1,'matter_intrinsic_power_red',iz,'red','solid')
#plot_pow_spec(ax1,'intrinsic_power',iz,'grey','dashed')
#plot_pow_spec(ax1,'intrinsic_power_blue',iz,'blue','dashed')
#plot_pow_spec(ax1,'intrinsic_power_red',iz,'red','dashed')

ax1.legend(loc='upper right',bbox_to_anchor=(1.2, 1.2),fontsize=8)
ax1.set_yscale('log')
ax1.set_xscale('log')
ax1.set_xlabel('k h/Mpc')
ax1.set_ylabel('P(k,z=0.75)')

# Now plot ratios to the NL matter power spectrum
pkvals_base = np.loadtxt(datadir+'matter_power_nl/p_k.txt')

#plot_ratio_pow_spec(ax2,'matter_intrinsic_power',pkvals_base,iz,'grey','solid')
#plot_ratio_pow_spec(ax2,'matter_intrinsic_power_blue',pkvals_base,iz,'blue','solid')
#plot_ratio_pow_spec(ax2,'matter_intrinsic_power_red',pkvals_base,iz,'red','solid')
#plot_ratio_pow_spec(ax2,'intrinsic_power',pkvals_base,iz,'grey','dashed')
#plot_ratio_pow_spec(ax2,'intrinsic_power_blue',pkvals_base,iz,'blue','dashed')
#plot_ratio_pow_spec(ax2,'intrinsic_power_red',pkvals_base,iz,'red','dashed')

ax2.set_xscale('log')
ax2.set_xlabel('k h/Mpc')
ax2.set_ylabel('P(k,z=0.75)/PNL(k,z=0.75)')


plt.tight_layout()
fig.savefig('3D_powerspectra_comparison.png')



"""A package for calculating the halo model."""

import contextlib
from importlib.metadata import PackageNotFoundError, version

with contextlib.suppress(PackageNotFoundError):
    __version__ = version(__name__)

from .add import UpsampledSpectra
from .bnl import NonLinearBias
from .hmi import CosmologyBase, HaloModelIngredients
from .hod import (
    HaloOccupationDistribution,
    Cacciato,
    Simple,
    Zehavi,
    Zhai,
    Zheng,
    load_data,
)
from .ia import AlignmentAmplitudes, SatelliteAlignment
from .pk import PowerSpectrumResult, Spectra
from .utils import poisson



import numpy as np
import pickle
import os
from onepower import Spectra
from scipy.integrate import trapezoid

from params_benchmark import kwargs


class OnepowerBenchmark:
    """
    Benchmark class for testing integration accuracy in power spectrum and HOD calculations.
    """

    def __init__(self, kwargs):
        """Initialize benchmark."""
        self.results = {}
        self.params = kwargs

    def _get_power_spectrum(self, spectra, spectrum_type):
        spectrum_methods = {
            'mm': spectra.power_spectrum_mm,
            'gg': spectra.power_spectrum_gg,
            'gm': spectra.power_spectrum_gm}
        return spectrum_methods[spectrum_type]

    def _test_convergence(self, quantity, values, spectra_types=['mm']):
        """
        Method to test convergence with different resolutions.

        Parameters
        ----------
        quantity : str
            Type of quantity being tested ('mass', 'k', 'nobs').
        values : list
            List of quantity values to test.
        spectra_types : list
            Types to compute. Either ['mm', 'gg', 'gm'], ['smf'], or ['hod'].

        Returns
        -------
        dict
            Results for each value and spectra type.
        """
        convergence_results = {}

        for value in values:
            modified_params = self._modify_params(self.params.copy(), quantity, value)
            spectra = Spectra(**modified_params)

            value_results = {}

            if 'smf' in spectra_types:
                value_results['smf'] = self._extract_smf_results(
                    spectra, modified_params)
            elif 'hod' in spectra_types:
                value_results['hod'] = self._extract_hod_results(
                    spectra, modified_params)
            else:
                # Power spectrum testing
                for spectrum_type in spectra_types:
                    value_results[spectrum_type] = self._extract_power_spectrum_results(
                        spectra, spectrum_type)

                    if quantity == 'mass':
                        value_results[spectrum_type].update({
                            'mass': spectra.mass,
                            'dlog10m': modified_params['dlog10m']})
                    elif quantity == 'k':
                        value_results[spectrum_type]['dlnk'] = modified_params['dlnk']

            convergence_results[value] = value_results

        return convergence_results

    def _extract_power_spectrum_results(self, spectra, spectrum_type):
        """
        Function to extract power spectrum results from objects.

        Parameters
        ----------
        spectra : Spectra object
            The spectra object containing power spectra.
        spectrum_type : str
            Type of spectrum ('mm', 'gg', 'gm').

        Returns
        -------
        dict
            Dictionary with extracted results.
        """
        ps_result = self._get_power_spectrum(spectra, spectrum_type)
        result = {
            'pk_tot': ps_result.pk_tot,
            'pk_1h': ps_result.pk_1h,
            'pk_2h': ps_result.pk_2h,
            'k_vec': spectra.k_vec}

        return result

    def _extract_smf_results(self, spectra, modified_params):
        """
        Function to extract SMF results from observable object.

        Parameters
        ----------
        spectra : Spectra object
            The spectra object containing observable.
        modified_params : dict
            Modified parameters used.

        Returns
        -------
        dict
            Dictionary with extracted results.
        """
        obs = spectra.obs

        result = {
            'stellar_mass': obs.obs[0, 0, 0, :],
            'smf_total': obs.obs_func[0, 0, :],
            'smf_cen': obs.obs_func_cen[0, 0, :],
            'smf_sat': obs.obs_func_sat[0, 0, :]}

        result['mass'] = spectra.mass
        result['dlog10m'] = modified_params['dlog10m']

        return result

    def _extract_hod_results(self, spectra, modified_params):
        """
        Function to extract HOD results from spectra object.

        Parameters
        ----------
        spectra : Spectra object
            The spectra object containing HOD.
        modified_params : dict
            Modified parameters used.

        Returns
        -------
        dict
            Dictionary with extracted results.
        """
        hod = spectra.hod

        result = {
            'mass': spectra.mass,
            'hod_cen': hod.hod_cen,
            'hod_sat': hod.hod_sat}

        result['nobs'] = modified_params['hod_settings']['nobs']

        return result

    def _modify_params(self, params, quantity, value):
        """
        Modify parameters based on quantity type.

        Parameters
        ----------
        params : dict
            Original parameters.
        quantity : str
            Type of quantity ('mass', 'k', or 'nobs').
        value : float or int
            Value to set.

        Returns
        -------
        dict
            Modified parameters.
        """
        if quantity == 'mass':
            params['dlog10m'] = value
        elif quantity == 'k':
            params['dlnk'] = value
        elif quantity == 'nobs':
            params['hod_settings']['nobs'] = value

        return params

    def create_benchmark(self, precision_params=None, filename=None):
        """
        Create high-resolution benchmark.

        Parameters
        ----------
        precision_params : dict
            Dictionary of precision parameters to override (e.g., {'dlnk': 0.000005, 'dlog10m': 0.005}).
            Must be provided.
        filename : str, optional
            Filename to save the benchmark to. If None, will auto-generate based on precision params.
        """
        if precision_params is None:
            raise ValueError("precision_params must be provided. Example: {'dlnk': 0.000005}")

        if filename is None:
            param_str = "_".join([f"{k}_{v}" for k, v in precision_params.items()])
            filename = f"high_resolution_benchmark_{param_str}.pkl"

        high_res_params = {**self.params, **precision_params}
        spectra = Spectra(**high_res_params)

        results_to_save = {
            'mm': spectra.power_spectrum_mm,
            'gg': spectra.power_spectrum_gg,
            'gm': spectra.power_spectrum_gm}

        self.save_to_file(high_res_params, results_to_save, spectra, filename)

        return filename

    def test_mass(self, dlog10m_values=[0.05, 0.01], spectra_types=['mm']):
        """Test power spectrum convergence with different mass array resolutions."""
        return self._test_convergence('mass', dlog10m_values, spectra_types)

    def test_k(self, dlnk_values=[0.001, 0.0001], spectra_types=['mm']):
        """Test power spectrum convergence with different k resolutions."""
        return self._test_convergence('k', dlnk_values, spectra_types)

    def test_mass_smf(self, dlog10m_values=[0.05, 0.005]):
        """Test SMF convergence with different halo mass resolutions."""
        return self._test_convergence('mass', dlog10m_values, spectra_types=['smf'])

    def test_nobs_hod(self, nobs_values=[300, 500]):
        """Test HOD convergence with different nobs resolutions."""
        return self._test_convergence('nobs', nobs_values, spectra_types=['hod'])

    def save_to_file(self, parameters, spectra_results, spectra, filename):
        """
        Save results to file.

        Parameters
        ----------
        parameters : dict
            Parameters used for computation.
        spectra_results : dict
            Dictionary with keys 'mm', 'gg', 'gm' containing results.
        spectra : object
            Spectra object containing k_vec, z_vec, mass.
        filename : str
            Filename to save to.
        """
        results = {
            'params': parameters,
            'spectra_results': {},
            'k_vec': spectra.k_vec,
            'z_vec': spectra.z_vec,
            'mass': spectra.mass}

        for spectrum_type, item in spectra_results.items():
            results['spectra_results'][spectrum_type] = {
                'pk_tot': item.pk_tot,
                'pk_1h': item.pk_1h,
                'pk_2h': item.pk_2h}

        with open(filename, 'wb') as f:
            pickle.dump(results, f)

        return


class SpectraWithTrapezoid(Spectra):
    """Spectra class that uses trapezoid integration instead of simpson."""

    def _replace_integration_method(self, integrand, x):
        """Helper method to use trapezoid instead of simpson."""
        return trapezoid(integrand, x, axis=-1)

    def compute_1h_term(self, profile_u, profile_v, mass, dndlnm):
        """Override to use trapezoid instead of simpson."""
        integrand = profile_u * profile_v * dndlnm / mass
        return self._replace_integration_method(integrand, mass)

    def compute_Im_term(self, mass, u_dm, b_dm, dndlnm, mean_density0):
        """Override to use trapezoid instead of simpson."""
        integrand_m = b_dm * dndlnm * u_dm * (1. / mean_density0)
        return self._replace_integration_method(integrand_m, mass)

    def compute_A_term(self, mass, b_dm, dndlnm, mean_density0):
        """Override to use trapezoid instead of simpson."""
        integrand_m1 = b_dm * dndlnm * (1.0 / mean_density0)
        A = 1.0 - self._replace_integration_method(integrand_m1, mass)
        if (A < 0.0).any():
            warnings.warn(
                'Warning: Mass function/bias correction is negative!',
                RuntimeWarning)
        return A

    def compute_Ig_term(self, profile, mass, dndlnm, b_m):
        """Override to use trapezoid instead of simpson."""
        integrand = profile * b_m * dndlnm / mass
        return self._replace_integration_method(integrand, mass)



"""
A module for computing intrinsic alignment properties.
This module provides classes and functions to calculate various properties
related to the intrinsic alignment of central and satellite galaxies within dark matter halos.
"""

import numpy as np
from astropy.io import fits
from hankel import HankelTransform
from scipy.fft import fht, fhtoffset
from scipy.integrate import simpson
from scipy.interpolate import RegularGridInterpolator, interp1d
from scipy.special import binom, gamma

from hmf._internals._cache import cached_quantity, parameter
from hmf._internals._framework import Framework


class AlignmentAmplitudes(Framework):
    """
    Class to IA amplitudes given options and dependencies on either luminosity, stellar mass, or halo mass.

    Parameters:
    -----------
    z_vec : array-like
        Redshift vector.
    central_ia_depends_on : str
        Dependency for central intrinsic alignment. Options: 'constant', 'luminosity', 'halo_mass'.
    satellite_ia_depends_on : str
        Dependency for satellite intrinsic alignment. Options: 'constant', 'luminosity', 'halo_mass'.
    gamma_2h_amplitude : float
        Amplitude for 2-halo term.
    beta_cen : float
        Beta parameter for central galaxies.
    beta_two : float, optional
        Secondary beta parameter.
    gamma_1h_amplitude : float
        Amplitude for 1-halo term.
    gamma_1h_slope : float
        Slope for 1-halo term.
    beta_sat : float
        Beta parameter for satellite galaxies.
    mpivot_cen : float, optional
        Pivot mass for central galaxies.
    mpivot_sat : float, optional
        Pivot mass for satellite galaxies.
    lpivot_cen : float, optional
        Pivot luminosity for central galaxies.
    lpivot_sat : float, optional
        Pivot luminosity for satellite galaxies.
    z_loglum_file_centrals : str, optional
        File path for central galaxy luminosity data.
    z_loglum_file_satellites : str, optional
        File path for satellite galaxy luminosity data.
    """

    def __init__(
        self,
        z_vec=None,
        central_ia_depends_on='halo_mass',
        satellite_ia_depends_on='halo_mass',
        gamma_2h_amplitude=5.33,
        beta_cen=0.44,
        beta_two=None,
        gamma_1h_amplitude=0.0015,
        gamma_1h_slope=-2.0,
        beta_sat=0.44,
        mpivot_cen=13.5,
        mpivot_sat=13.5,
        lpivot_cen=None,
        lpivot_sat=None,
        z_loglum_file_centrals=None,
        z_loglum_file_satellites=None,
    ):
        self.z_vec = z_vec
        self.central_ia_depends_on = central_ia_depends_on
        self.satellite_ia_depends_on = satellite_ia_depends_on

        self.gamma_2h_amplitude = gamma_2h_amplitude
        self.beta_cen = beta_cen
        self.beta_two = beta_two
        self.gamma_1h_slope = gamma_1h_slope
        self.gamma_1h_amplitude = gamma_1h_amplitude
        self.beta_sat = beta_sat

        self.mpivot_cen = mpivot_cen
        self.mpivot_sat = mpivot_sat
        self.lpivot_cen = lpivot_cen
        self.lpivot_sat = lpivot_sat

        self.z_loglum_file_centrals = z_loglum_file_centrals
        self.z_loglum_file_satellites = z_loglum_file_satellites

    @parameter('param')
    def z_vec(self, val):
        """
        Redshift vector.

        :type: array-like
        """
        return val

    @parameter('param')
    def central_ia_depends_on(self, val):
        """
        Validate the central intrinsic alignment dependencies.

        Raises:
        -------
        ValueError : If an invalid option is provided for central_ia_depends_on or satellite_ia_depends_on.
        """
        valid_options = ['constant', 'luminosity', 'halo_mass']
        if val not in valid_options:
            raise ValueError(
                f'Choose one of the following options for central_IA_depends_on: {valid_options}'
            )
        return val

    @parameter('param')
    def satellite_ia_depends_on(self, val):
        """
        Validate the satellite intrinsic alignment dependencies.

        Raises:
        -------
        ValueError : If an invalid option is provided for central_ia_depends_on or satellite_ia_depends_on.
        """
        valid_options = ['constant', 'luminosity', 'halo_mass']
        if val not in valid_options:
            raise ValueError(
                f'Choose one of the following options for satellite_IA_depends_on: {valid_options}'
            )
        return val

    @parameter('param')
    def gamma_2h_amplitude(self, val):
        """
        Amplitude for 2-halo term.

        :type: float
        """
        return val

    @parameter('param')
    def gamma_1h_slope(self, val):
        """
        Slope for 1-halo term.

        :type: float
        """
        return val

    @parameter('param')
    def gamma_1h_amplitude(self, val):
        """
        Amplitude for 1-halo term.

        :type: float
        """
        return val

    @parameter('param')
    def beta_cen(self, val):
        """
        Beta parameter for central galaxies.

        :type: float
        """
        return val

    @parameter('param')
    def beta_sat(self, val):
        """
        beta_sat : float
            Beta parameter for satellite galaxies.
        """
        return val

    @parameter('param')
    def mpivot_cen(self, val):
        """
        Pivot mass for central galaxies.

        :type: float
        """
        return 10.0**val if val is not None else None

    @parameter('param')
    def mpivot_sat(self, val):
        """
        Pivot mass for satellite galaxies.

        :type: float
        """
        return 10.0**val if val is not None else None

    @parameter('param')
    def lpivot_cen(self, val):
        """
        Pivot luminosity for central galaxies.

        :type: float
        """
        return 10.0**val if val is not None else None

    @parameter('param')
    def lpivot_sat(self, val):
        """
        Pivot luminosity for satellite galaxies.

        :type: float
        """
        return 10.0**val if val is not None else None

    @parameter('param')
    def z_loglum_file_centrals(self, val):
        """
        File path for central galaxy luminosity data.

        :type: str
        """
        return val

    @parameter('param')
    def z_loglum_file_satellites(self, val):
        """
        File path for satellite galaxy luminosity data.

        :type: str
        """
        return val

    @cached_quantity
    def lum_centrals(self):
        """
        Returns the luminosity array for centrals.
        """
        return self._initialize_luminosity_array('centrals')[0]

    @cached_quantity
    def lum_pdf_z_centrals(self):
        """
        Returns the luminosity array for satellites.
        """
        return self._initialize_luminosity_array('centrals')[1]

    @cached_quantity
    def lum_satellites(self):
        """
        Returns the luminosity PDF array for centrals.
        """
        return self._initialize_luminosity_array('satellites')[0]

    @cached_quantity
    def lum_pdf_z_satellites(self):
        """
        Returns the luminosity PDF array for satellites.
        """
        return self._initialize_luminosity_array('satellites')[1]

    def _initialize_luminosity_array(self, galaxy_type):
        """
        Initialize and return the luminosity array based on galaxy type.

        Parameters:
        -----------
        galaxy_type : str
            Type of galaxy, either 'centrals' or 'satellites'.

        Returns:
        --------
        numpy.ndarray : lum
            Luminosity array.
        """
        z_loglum_file = (
            self.z_loglum_file_centrals
            if galaxy_type == 'centrals'
            else self.z_loglum_file_satellites
        )
        if z_loglum_file is None:
            raise ValueError(
                f'You have not provided a luminosity file for {galaxy_type}. Please include z_loglum_file_{galaxy_type}.'
            )

        nlbins = 10000
        lum, lum_pdf_z = self.compute_luminosity_pdf(z_loglum_file, nlbins)

        return lum, lum_pdf_z

    def mean_l_l0_to_beta(self, xlum, pdf, l0, beta):
        """
        Compute the mean luminosity scaling.

        Parameters:
        -----------
        xlum : array-like
            Luminosity values.
        pdf : array-like
            Probability density function values.
        l0 : float
            Pivot luminosity.
        beta : float
            Beta parameter.

        Returns:
        --------
        float : Mean luminosity scaling.
        """
        return simpson(pdf * (xlum / l0) ** beta, x=xlum)

    def broken_powerlaw(self, xlum, pdf, gamma_2h_lum, l0, beta, beta_low):
        """
        Compute the broken power law.

        Parameters:
        -----------
        xlum : array-like
            Luminosity values.
        pdf : array-like
            Probability density function values.
        gamma_2h_lum : float
            Amplitude for 2-halo term.
        l0 : float
            Pivot luminosity.
        beta : float
            Beta parameter.
        beta_low : float
            Secondary beta parameter.

        Returns:
        --------
        float : Integral of the broken power law.
        """
        alignment_ampl = np.where(
            xlum > l0,
            gamma_2h_lum * (xlum / l0) ** beta,
            gamma_2h_lum * (xlum / l0) ** beta_low,
        )
        return simpson(pdf * alignment_ampl, x=xlum)

    def compute_luminosity_pdf(self, z_loglum_file, nlbins):
        """
        Compute the luminosity PDF.

        Parameters:
        -----------
        z_loglum_file : str
            File path for galaxy luminosity data.
        nlbins : int
            Number of bins for the histogram.

        Returns:
        --------
        tuple : (bincen, pdf)
            Bin centers and probability density function values.
        """
        galfile = fits.open(z_loglum_file)[1].data
        z_gal = np.array(galfile['z'])
        loglum_gal = np.array(galfile['loglum'])

        z_bins = self.z_vec
        dz = 0.5 * (z_bins[1] - z_bins[0])
        z_edges = np.append(z_bins - dz, z_bins[-1] + dz)

        bincen = np.zeros([self.z_vec.size, nlbins])
        pdf = np.zeros([self.z_vec.size, nlbins])

        for i in range(self.z_vec.size):
            mask_z = (z_gal >= z_edges[i]) & (z_gal < z_edges[i + 1])
            loglum_bin = loglum_gal[mask_z]
            if loglum_bin.size:
                lum = 10.0**loglum_bin
                pdf_tmp, _lum_bins = np.histogram(lum, bins=nlbins, density=True)
                _dbin = (_lum_bins[-1] - _lum_bins[0]) / (1.0 * nlbins)
                bincen[i] = _lum_bins[:-1] + 0.5 * _dbin
                pdf[i] = pdf_tmp

        return bincen, pdf

    @cached_quantity
    def alignment_gi(self):
        """
        Compute the alignment_gi based on the central galaxies' properties.
        """
        if self.central_ia_depends_on == 'constant':
            return self.gamma_2h_amplitude * np.ones_like(self.z_vec)
        elif self.central_ia_depends_on == 'luminosity':
            if self.lpivot_cen is None:
                raise ValueError(
                    'You have chosen central luminosity scaling without providing a pivot luminosity parameter. Include lpivot_cen.'
                )
            if self.beta_two is not None:
                mean_lscaling = np.array(
                    [
                        self.broken_powerlaw(
                            self.lum_centrals[i],
                            self.lum_pdf_z_centrals[i],
                            self.gamma_2h_amplitude,
                            self.lpivot_cen,
                            self.beta_cen,
                            self.beta_two,
                        )
                        for i in range(self.z_vec.size)
                    ]
                )
            else:
                mean_lscaling = self.gamma_2h_amplitude * self.mean_l_l0_to_beta(
                    self.lum_centrals,
                    self.lum_pdf_z_centrals,
                    self.lpivot_cen,
                    self.beta_cen,
                )
            return mean_lscaling
        elif self.central_ia_depends_on == 'halo_mass':
            if self.mpivot_cen is None:
                raise ValueError(
                    'You have chosen central halo-mass scaling without providing a pivot mass parameter. Include mpivot_cen.'
                )
            if self.beta_two is not None:
                raise ValueError(
                    'A double power law model for the halo mass dependence of centrals has not been implemented.'
                )
            return self.gamma_2h_amplitude * np.ones_like(self.z_vec)

    @cached_quantity
    def gamma_1h_amp(self):
        """
        Compute the gamma_1h_amplitude based on the satellite galaxies' properties.
        """
        if self.satellite_ia_depends_on == 'constant':
            return self.gamma_1h_amplitude * np.ones_like(self.z_vec)
        elif self.satellite_ia_depends_on == 'luminosity':
            if self.lpivot_sat is None:
                raise ValueError(
                    'You have chosen satellite luminosity scaling without providing a pivot luminosity parameter. Include lpivot_sat.'
                )
            mean_lscaling = self.mean_l_l0_to_beta(
                self.lum_satellites,
                self.lum_pdf_z_satellites,
                self.lpivot_sat,
                self.beta_sat,
            )
            return self.gamma_1h_amplitude * mean_lscaling
        elif self.satellite_ia_depends_on == 'halo_mass':
            if self.mpivot_sat is None:
                raise ValueError(
                    'You have chosen satellite halo-mass scaling without providing a pivot mass parameter. Include mpivot_sat.'
                )
            return self.gamma_1h_amplitude * np.ones_like(self.z_vec)


class SatelliteAlignment(AlignmentAmplitudes):
    """
    A class to compute the alignment properties of satellite galaxies within dark matter halos.
    This includes calculating the Hankel transform, radial profiles, and other related quantities.

    Parameters:
    -----------
    mass_in : array_like
        Array of halo masses.
    c_in : array_like
        Concentration parameter.
    r_s_in : array_like
        Scale radius.
    rvir_in : array_like
        Virial radius.
    n_hankel : int, optional
        Number of steps in the Hankel transform integration.
    nmass : int, optional
        Number of mass bins.
    nk : int, optional
        Number of k bins.
    ell_max : int, optional
        Maximum multipole moment.
    truncate : bool, optional
        Whether to truncate the NFW profile at the virial radius.
    method : str, optional
        Which method to perform Fourier/Hankel transform to use
    amplitude_kwargs : dict
        Extra parameters passed to the AlignmentAmplitudes class.
    """

    def __init__(
        self,
        mass_in=None,
        c_in=None,
        r_s_in=None,
        rvir_in=None,
        n_hankel=350,
        nmass=5,
        nk=10,
        ell_max=6,
        truncate=False,
        method='fftlog',
        **amplitude_kwargs,
    ):
        # Call super init MUST BE DONE FIRST.
        super().__init__(**amplitude_kwargs)

        self.method = method
        self.nmass = nmass
        self.ell_max = ell_max
        self.truncate = truncate
        self.nk = nk
        self.n_hankel = n_hankel
        self.mass_in = mass_in
        self.c_in = c_in
        self.r_s_in = r_s_in
        self.rvir_in = rvir_in

        # These are for now hardcoded choices
        self.theta_k = np.pi / 2.0
        self.phi_k = 0.0

    @parameter('param')
    def mass_in(self, val):
        """
        Array of halo masses.

        :type: array_like
        """
        return val

    @parameter('param')
    def c_in(self, val):
        """
        Concentration parameter.

        :type: array_like
        """
        return val

    @parameter('param')
    def r_s_in(self, val):
        """
        Scale radius.

        :type: array_like
        """
        return val

    @parameter('param')
    def rvir_in(self, val):
        """
        Virial radius.

        :type: array_like
        """
        return val

    @parameter('param')
    def n_hankel(self, val):
        """
        Number of steps in the Hankel transform integration.

        :type: int
        """
        return val

    @parameter('param')
    def nmass(self, val):
        """
        Number of mass bins.

        :type: int
        """
        return int(val)

    @parameter('param')
    def nk(self, val):
        """
        Number of k bins.

        :type: int
        """
        return val

    @parameter('param')
    def ell_max(self, val):
        """
        Maximum multipole moment.

        :type: int
        """
        if val > 11:
            raise ValueError(
                'Please reduce ell_max < 11 or update ia_radial_interface.py'
            )
        return val

    @parameter('param')
    def truncate(self, val):
        """
        Whether to truncate the NFW profile at the virial radius.

        :type: bool
        """
        return val

    @parameter('param')
    def method(self, val):
        """
        Which method to perform Fourier/Hankel transform to use

        :type: str
        """
        valid_methods = ['hankel', 'fftlog']
        if val not in valid_methods:
            raise ValueError(
                f'The valid methods to evaluate the fourier tranform of IA shear field are: {valid_methods}. Requested method: {val}!'
            )
        return val

    @cached_quantity
    def ell_values(self):
        """
        Sets the ell values array.

        Returns:
        --------
        array_like
            ell_values
        """
        # CCL and Fortuna use ell_max=6. SB10 uses ell_max = 2.
        # Higher/lower increases/decreases accuracy but slows/speeds the code
        return np.arange(0, self.ell_max + 1, 2)

    @cached_quantity
    def k_vec(self):
        """
        Sets the k vector array.

        Returns:
        --------
        array_like
            k_vec
        """
        nk = 100 if self.method == 'fftlog' else self.nk
        return np.logspace(np.log10(1e-3), np.log10(1e3), nk)

    def _downsample(self, quantity):
        """
        Downsample the halo parameters to reduce computational complexity.

        Parameter:
        -----------
        quantity : array_like
            An array with a quantity that needs to be downsamples (c, rs, rvir or mass)

        Returns:
        --------
        array
            Downsampled arrays of either halo masses, concentration parameters, scale radii, and virial radii.
        """
        if self.method != 'hankel':
            return quantity

        if self.mass_in.size == self.nmass:
            return quantity

        if self.mass_in.size < self.nmass:
            raise ValueError(
                'The halo mass resolution is too low for the radial IA calculation. '
                'Please increase nmass when you run halo_model_ingredients.py'
            )

        downsample_factor = self.mass_in.size // self.nmass

        if isinstance(quantity, np.ndarray) and quantity.ndim == 1:
            downsampled_quantity = quantity[::downsample_factor]
            if downsampled_quantity[-1] != quantity[-1]:
                downsampled_quantity = np.append(downsampled_quantity, quantity[-1])
        else:
            downsampled_quantity = quantity[:, ::downsample_factor]
            if np.all(downsampled_quantity[:, -1] != quantity[:, -1]):
                downsampled_quantity = np.concatenate(
                    (downsampled_quantity, np.atleast_2d(quantity[:, -1]).T), axis=1
                )
        return downsampled_quantity

    @cached_quantity
    def mass(self):
        """
        Downsampled mass_in if method == hankel, mass_in otherwise.

        Returns:
        --------
        array_like
            mass
        """
        return self._downsample(self.mass_in)

    @cached_quantity
    def c(self):
        """
        Downsampled c_in if method == hankel, c_in otherwise.

        Returns:
        --------
        array_like
            c
        """
        return self._downsample(self.c_in)

    @cached_quantity
    def r_s(self):
        """
        Downsampled r_s_in if method == hankel, r_s_in otherwise.

        Returns:
        --------
        array_like
            r_s
        """
        return self._downsample(self.r_s_in)

    @cached_quantity
    def rvir(self):
        """
        Downsampled rvir_in if method == hankel, rvir_in otherwise.

        Returns:
        --------
        array_like
            rvir
        """
        return self._downsample(self.rvir_in)

    @cached_quantity
    def hankel(self):
        """
        Initialize Hankel transform.

        We've used hankel.get_h to set h, N is then h=pi/N, finding best_h = 0.05, best_N=62
        If you want perfect agreement with CCL use: N=50000, h=0.00006 (VERY SLOW!!)

        Ideally we need to find a way to just evaluate this part outside of the class
        so it can work nicely with CosmoSIS setup function

        Returns:
        --------
        list
            A list of HankelTransform objects for each multipole moment.
        """
        if self.method == 'hankel':
            return [
                HankelTransform(ell + 0.5, self.n_hankel, np.pi / self.n_hankel)
                for ell in self.ell_values
            ]
        else:
            return None

    def I_x(self, a, b):
        """
        Compute the integral of (1 - x^2)^(a/2) * x^b from -1 to 1.

        Parameters:
        -----------
        a : float
            Exponent for the (1 - x^2) term.
        b : float
            Exponent for the x term.

        Returns:
        --------
        float
            The value of the integral.
        """
        return (
            (1.0 + (-1.0) ** b)
            * gamma(a / 2.0 + 1.0)
            * gamma((b + 1.0) / 2.0)
            / (2.0 * gamma(a / 2.0 + b / 2.0 + 3.0 / 2.0))
        )

    def calculate_f_ell(self, ell, gamma_b):
        """
        Computes the angular part of the satellite intrinsic shear field.

        Eq. (C8) in `Fortuna et al. 2021 <https://arxiv.org/abs/2003.02700>`

        Parameters:
        -----------
        ell : int
            Multipole moment.
        gamma_b : float
            Slope parameter.

        Returns:
        --------
        complex
            The value of the angular part of the satellite intrinsic shear field.
        """
        phase = np.cos(2.0 * self.phi_k) + 1j * np.sin(2.0 * self.phi_k)

        # Follow CCL by hard-coding for most common cases (b=0, b=-2) to gain speed
        # (in CCL gain is ~1.3sec - gain here depends on how many times this is called).
        if self.theta_k == np.pi / 2.0 and gamma_b in [0, -2]:
            pre_calc_f_ell = {
                0: np.array(
                    [
                        0,
                        0,
                        2.77582637,
                        0,
                        -0.19276603,
                        0,
                        0.04743899,
                        0,
                        -0.01779024,
                        0,
                        0.00832446,
                        0,
                        -0.00447308,
                        0,
                    ]
                ),
                -2: np.array(
                    [
                        0,
                        0,
                        4.71238898,
                        0,
                        -2.61799389,
                        0,
                        2.06167032,
                        0,
                        -1.76714666,
                        0,
                        1.57488973,
                        0,
                        -1.43581368,
                        0,
                    ]
                ),
            }
            return pre_calc_f_ell.get(gamma_b)[ell] * phase

        else:
            # If either of the above expressions are met the return statement is executed and the function ends.
            # Otherwise, the function continues to calculate the general case.
            gj = np.array(
                [
                    0,
                    0,
                    np.pi / 2,
                    0,
                    np.pi / 2,
                    0,
                    15 * np.pi / 32,
                    0,
                    7 * np.pi / 16,
                    0,
                    105 * np.pi / 256,
                    0,
                ]
            )
            sum1 = sum(
                binom(ell, m)
                * binom(0.5 * (ell + m - 1.0), ell)
                * sum(
                    binom(m, j)
                    * gj[j]
                    * np.sin(self.theta_k) ** j
                    * np.cos(self.theta_k) ** (m - j)
                    * self.I_x(j + gamma_b, m - j)
                    for j in range(m + 1)
                )
                for m in range(ell + 1)
            )
            return 2.0**ell * sum1 * phase

    @cached_quantity
    def wkm_f_ell(self):
        """
        Integral of the angular part in eq B8 (SB10) using the Legendre polynomials assuming theta_e=theta, phi_e=phi (perfect radial alignment).

        Note CCL only calculates the real parts of w(k|m)f_ell and doesn't take the absolute value....
        which means you'll get negative values for wkm in CCL: they take the absolute value later.

        Returns:
        --------
        ndarray
            The absolute value of the sum of the radial and angular parts.
        """
        uell = self.compute_uell_gamma_r_hankel
        nz, nm, nk = uell.shape[1], uell.shape[2], uell.shape[3]
        sum_ell = np.zeros([nz, nm, nk], dtype=complex)

        for ell in self.ell_values:
            angular = self.calculate_f_ell(ell, self.gamma_1h_slope)
            radial = (1j) ** ell * (2.0 * ell + 1.0) * uell[ell // 2, :, :, :]
            sum_ell += radial * angular

        return np.abs(sum_ell)

    def gamma_r_nfw_profile(self, r, rs, rvir, a, b, rcore=0.06, truncate=True):
        """
        Compute the radial profile of the NFW (Navarro-Frenk-White) profile with a power-law correction.

        Parameters:
        -----------
        r : float
            Radial distance.
        rs : float
            Scale radius.
        rvir : float
            Virial radius.
        a : float
            Amplitude of the power-law correction.
        b : float
            Slope of the power-law correction.
        rcore : float, optional
            Core radius.
        truncate : bool, optional
            Whether to truncate the profile at the virial radius.

        Returns:
        --------
        float
            The value of the radial profile.
        """
        gamma = a * (r / rvir) ** b
        gamma = np.where(r < rcore, a * (rcore / rvir) ** b, gamma)
        gamma = np.clip(gamma, None, 0.3)

        nfw = 1.0 / ((r / rs) * (1.0 + (r / rs)) ** 2.0)
        if truncate:
            nfw = np.where(r >= rvir, 0.0, nfw)
        return gamma * nfw

    @cached_quantity
    def compute_uell_gamma_r_hankel(self):
        """
        Computes a 4D array containing u_ell as a function of l, z, m, and k. THIS FUNCTION IS THE SLOWEST PART!

        h_transf = HankelTransform(ell+0.5,N_hankel,pi/N_hankel)
        Note even though ell is not used in this function, h_transf depends on ell
        We initialize the class in setup as it only depends on predefined ell values

        Note: I experimented coding the use of Simpson integration for where the Bessel function is flat
        and then switching to the Hankel transform for where the Bessel function oscillates.
        This is more accurate than using the Hankel transform for all k values with lower accuracy
        settings, but it's slower than using the Hankel transform for all k values.
        It's also difficult to decide how to define the transition between the two methods.
        Given that low-k accuracy is unimportant for IA, I've decided to use the Hankel transform for all k values.

        Returns:
        --------
        ndarray
            A 4D array of u_ell values.
        """
        mnfw = (
            4.0
            * np.pi
            * self.r_s**3.0
            * (np.log(1.0 + self.c) - self.c / (1.0 + self.c))
        )
        uk_l = np.zeros(
            [self.ell_values.size, self.z_vec.size, self.mass.size, self.k_vec.size]
        )

        for i, ell in enumerate(self.ell_values):
            if self.method == 'hankel':
                for jz in range(self.z_vec.size):
                    for im in range(self.mass.size):
                        nfw_f = lambda x: self.gamma_r_nfw_profile(
                            x,
                            self.r_s[jz, im],
                            self.rvir[im],
                            self.gamma_1h_amp[jz],
                            self.gamma_1h_slope,
                            truncate=self.truncate,
                        ) * np.sqrt((x * np.pi) / 2.0)
                        uk_l[i, jz, im, :] = self.hankel[i].transform(
                            nfw_f, self.k_vec
                        )[0] / (self.k_vec**0.5 * mnfw[jz, im])

            if self.method == 'fftlog':
                mu = ell + 0.5
                # Precision settings that seem to be working fine
                bias = -0.5 * (i + 2.0)
                low_r = -8
                high_r = 8
                r = np.logspace(low_r, high_r, 512)
                dlnr = np.log(r[1] / r[0])
                offset = fhtoffset(dlnr, mu=mu, initial=-2 * np.log(10.0), bias=bias)
                k = np.exp(offset) / r[::-1]

                nfw_f = (
                    self.gamma_r_nfw_profile(
                        r[np.newaxis, np.newaxis, :],
                        self.r_s[:, :, np.newaxis],
                        self.rvir[np.newaxis, :, np.newaxis],
                        self.gamma_1h_amp[:, np.newaxis, np.newaxis],
                        self.gamma_1h_slope,
                        truncate=self.truncate,
                    )
                    * r**1.5
                    * np.sqrt(np.pi / 2.0)
                )
                ft = fht(nfw_f, dlnr, mu=mu, offset=offset, bias=bias) / (
                    k**1.5 * mnfw[:, :, np.newaxis]
                )
                uk_l[i, :, :, :] = interp1d(
                    k,
                    ft,
                    axis=-1,
                    kind='linear',
                    fill_value='extrapolate',
                    bounds_error=False,
                )(self.k_vec)

        return uk_l

    @cached_quantity
    def wkm(self):
        """
        Return the computed wkm_f_ell values along with the redshift, mass, and k vectors.

        Returns:
        --------
        tuple
            Computed values of wkm_f_ell, array of redshifts, array of halo masses, and array of k values.
        """
        return self.wkm_f_ell, self.z_vec, self.mass, self.k_vec

    def upsampled_wkm(self, k_vec_out, mass_out):
        """
        Interpolates the wkm profiles and upsamples back to the original grid.

        Parameters:
        -----------
        k_vec_out : array_like
            Output array of k values.
        mass_out : array_like
            Output array of halo masses.

        Returns:
        --------
        ndarray
            Upsampled array of wkm values.
        """
        wkm_in = self.wkm_f_ell
        if self.method == 'hankel':
            wkm_out = np.empty([self.z_vec.size, mass_out.size, k_vec_out.size])
            for jz in range(self.z_vec.size):
                # Create the interpolator
                lg_w_interp2d = RegularGridInterpolator(
                    (np.log10(self.k_vec).T, np.log10(self.mass).T),
                    np.log10(wkm_in[jz, :, :] / self.k_vec**2).T,
                    bounds_error=False,
                    fill_value=None,
                )

                # Prepare the grid for interpolation
                lgkk, lgmm = np.meshgrid(
                    np.log10(k_vec_out), np.log10(mass_out), sparse=True
                )

                # Interpolate the values
                lg_wkm_interpolated = lg_w_interp2d((lgkk.T, lgmm.T)).T

                # Convert back to original scale
                wkm_out[jz, :, :] = 10.0 ** (lg_wkm_interpolated) * k_vec_out**2.0
            return wkm_out

        if self.method == 'fftlog':
            # Need to only upsample the k vector!
            return interp1d(
                self.k_vec,
                wkm_in,
                axis=-1,
                kind='linear',
                fill_value='extrapolate',
                bounds_error=False,
            )(k_vec_out)



import configparser as cfp
import numpy as np
import twopoint
import wrapper_twopoint as wtp
from cosmosis.datablock import option_section
from cosmosis.datablock.cosmosis_py import errors

# New scale cuts supporting 1pt statistics

def setup(options):
    config = {}

    ## Read inputs from datablock
    config['output_section_name'] = options.get_string(option_section, 'output_section_name', default='likelihood')
    try:
        config['data_and_covariance_fits_filename'] = options.get_string(option_section, 'data_and_covariance_fits_filename')
    except errors.BlockNameNotFound:
        raise NameError('data_and_covariance_fits_filename cannot be empty')

    ## Read extension names for data outputs
    config['wt_extension_name']                       = options.get_string(option_section, 'wt_extension_name', default='wtheta')
    config['gt_extension_name']                       = options.get_string(option_section, 'gt_extension_name', default='gammat')
    config['xi_plus_extension_name']                  = options.get_string(option_section, 'xi_plus_extension_name', default='xip')
    config['xi_minus_extension_name']                 = options.get_string(option_section, 'xi_minus_extension_name', default='xim')
    config['bandpower_clustering_extension_name']     = options.get_string(option_section, 'bandpower_clustering_extension_name', default='Pnn')
    config['bandpower_ggl_extension_name']            = options.get_string(option_section, 'bandpower_ggl_extension_name', default='PneE')
    config['bandpower_e_cosmic_shear_extension_name'] = options.get_string(option_section, 'bandpower_e_cosmic_shear_extension_name', default='PeeE')
    config['bandpower_b_cosmic_shear_extension_name'] = options.get_string(option_section, 'bandpower_b_cosmic_shear_extension_name', default='PeeB')
    config['psi_stats_gg_extension_name']             = options.get_string(option_section, 'psi_stats_gg_extension_name', default='Psi_gg')
    config['psi_stats_gm_extension_name']             = options.get_string(option_section, 'psi_stats_gm_extension_name', default='Psi_gm')
    config['cosebis_extension_name']                  = options.get_string(option_section, 'cosebis_extension_name', default='En')
    config['onepoint_extension_name']                 = options.get_string(option_section, 'onepoint_extension_name', default='1pt')

    ## Read section names for theory outputs
    config['wt_section_name']                       = options.get_string(option_section, 'wt_section_name', default='galaxy_xi')
    config['gt_section_name']                       = options.get_string(option_section, 'gt_section_name', default='galaxy_shear_xi')
    config['xi_plus_section_name']                  = options.get_string(option_section, 'xi_plus_section_name', default='shear_xi_plus')
    config['xi_minus_section_name']                 = options.get_string(option_section, 'xi_minus_section_name', default='shear_xi_minus')
    config['bandpower_clustering_section_name']     = options.get_string(option_section, 'bandpower_clustering_section_name', default='bandpower_clustering')
    config['bandpower_ggl_section_name']            = options.get_string(option_section, 'bandpower_ggl_section_name', default='bandpower_ggl')
    config['bandpower_e_cosmic_shear_section_name'] = options.get_string(option_section, 'bandpower_e_cosmic_shear_section_name', default='bandpower_e_cosmic_shear')
    config['bandpower_b_cosmic_shear_section_name'] = options.get_string(option_section, 'bandpower_b_cosmic_shear_section_name', default='bandpower_b_cosmic_shear')
    config['psi_stats_gg_section_name']             = options.get_string(option_section, 'psi_stats_gg_section_name', default='psi_gg')
    config['psi_stats_gm_section_name']             = options.get_string(option_section, 'psi_stats_gm_section_name', default='psi_gm')
    config['cosebis_section_name']                  = options.get_string(option_section, 'cosebis_section_name', default='cosebis')
    config['onepoint_section_name']                 = options.get_string(option_section, 'onepoint_section_name', default='observable_function')

    ## Read scale cuts
    try:
        ## Try to load scale cuts file if specified
        config['scale_cuts_filename'] = options.get_string(option_section, 'scale_cuts_filename')
        try:
            parser = cfp.ConfigParser()
            parser.read(config['scale_cuts_filename'])
        except:
            raise OSError('\"%s\" not found' % config['scale_cuts_filename'])

        config['scale_cuts_option'] = options.get_string(option_section, 'scale_cuts_option', default='scale_cuts_none')
        try:
            ## Make scale cuts arguments
            scDict = parser[config['scale_cuts_option']]
        except:
            raise KeyError('Bad scale cuts option: \"%s\"' % config['scale_cuts_option'])

    except errors.BlockNameNotFound:
        ## Otherwise read inline options
        key_list   = [k for _, k in options.keys(option_section) if k.split('_')[0] in ['use', 'cut', 'keep']]
        value_list = [str(options[option_section, k]).strip(' []') for k in key_list]
        scDict = {k : v for k, v in zip(key_list, value_list)}

    ## Load data & cov file
    try:
        TP_data = wtp.TwoPointWrapper.from_fits(config['data_and_covariance_fits_filename'], covmat_name='COVMAT')
    except:
        raise OSError('\"%s\" not found' % config['data_and_covariance_fits_filename'])

    labConv = wtp.LabelConvention(w=config['wt_extension_name'],
                                  gamma_t=config['gt_extension_name'],
                                  xi_p=config['xi_plus_extension_name'],
                                  xi_m=config['xi_minus_extension_name'],
                                  P_nn=config['bandpower_clustering_extension_name'],
                                  P_ne_E=config['bandpower_ggl_extension_name'],
                                  P_ee_E=config['bandpower_e_cosmic_shear_extension_name'],
                                  E_n=config['cosebis_extension_name'],
                                  Psi_gg=config['psi_stats_gg_extension_name'],
                                  Psi_gm=config['psi_stats_gm_extension_name'],
                                  onept=config['onepoint_extension_name'])
    statsList, scArgs = labConv.makeScaleCutsArgs(scDict) ## Here, we convert the keys from the default ones to the custom ones.
    config['scale_cuts_arguments'] = scArgs
    config['label_convention']     = labConv

    ## Do scale cuts to data and cov
    TP_data.cutScales(cutCross=scArgs[0], statsTag_tomoInd_tomoInd_list=scArgs[1], statsTag_binIndList_dict=scArgs[2])
    TP_data.keepScales(statsTag_tomoInd1_tomoInd2__angMin_angMax_dict=scArgs[3], statsTag__angMin_angMax_dict=scArgs[4])
    #print('  Did scale cuts to data & cov')

    ## Don't put this before scale cuts, because TP_data.choose_data_sets does
    ## not modify the covmat_info attribute but only the covmat attribute.
    ## While one can cut a correctly cut covariance, the scale cuts on this
    ## covariance, which relies on covmat_info, can result in errors.
    statsList_c           = [labConv.defaultToCustomStatsTag(stats) for stats in statsList]
    config['use_stats']   = statsList
    config['use_stats_c'] = statsList_c
    TP_data.choose_data_sets(statsList_c)

    ## Extract the vector & matrix & put in config dict
    config['data']       = TP_data.makeMeanVector()
    config['covariance'] = TP_data.covmat
    config['inv_covariance'] = np.linalg.inv(TP_data.covmat)

    config['simulate'] = options.get_bool(option_section, 'simulate', default=False)
    config['simulate_with_noise'] = options.get_bool(option_section, 'simulate_with_noise', default=True)
    config['number_of_simulations'] = options.get_int(option_section,'number_of_simulations',default=1)
    config['mock_filename'] = options.get_string(option_section, 'mock_filename', default="")
    if config['simulate']:
        config["TP_data"] = TP_data

    return config

def execute(block, config):
    #print('Gathering theory outputs to make a vector')

    ## Save data and cov in the data block
    output_section_name = config['output_section_name']
    block[output_section_name, 'data']       = config['data']
    block[output_section_name, 'covariance'] = config['covariance']
    block[output_section_name, 'inv_covariance'] = config['inv_covariance']

    ## Define some things first
    labConv    = config['label_convention']
    nbTomo_max = 1000
    spectra    = []

    ## Don't change the order of this list
    ## Read as: [section_name, extension_name, angle_name, isGGL]
    sectionNameList = [
        [config['wt_section_name'],                       config['wt_extension_name'],                       'theta',         False],
        [config['gt_section_name'],                       config['gt_extension_name'],                       'theta',         True],
        [config['xi_plus_section_name'],                  config['xi_plus_extension_name'],                  'theta',         False],
        [config['xi_minus_section_name'],                 config['xi_minus_extension_name'],                 'theta',         False],
        [config['bandpower_clustering_section_name'],     config['bandpower_clustering_extension_name'],     'ell',           False],
        [config['bandpower_ggl_section_name'],            config['bandpower_ggl_extension_name'],            'ell',           True],
        [config['bandpower_e_cosmic_shear_section_name'], config['bandpower_e_cosmic_shear_extension_name'], 'ell',           False],
        [config['psi_stats_gg_section_name'],             config['psi_stats_gg_extension_name'],             'index_bin_1_1',     False],
        [config['psi_stats_gm_section_name'],             config['psi_stats_gm_extension_name'],             'index_bin_1_1',     True],
        [config['cosebis_section_name'],                  config['cosebis_extension_name'],                  'n',     False],
    ]

    for line in sectionNameList:
        section_name, extension_name, angle_name, isGGL = line

        if extension_name not in config['use_stats']:
            #print('  Skipped %s' % extension_name)
            continue

        elif not block.has_section(section_name):
            raise AssertionError('You specified \"{}\" in use_stats. When I was computing for \"{}\", I tried to\n\
                grab theory values from \"{}\" but could not find anything.'.format(config['use_stats'], extension_name, section_name))

        ## Initialize!
        sBuilder = wtp.SpectrumBuilder()

        ## Is it BP? If yes, then ell needs to be calculated instead of calling directly from data block.
        if angle_name == 'ell':
            lower = block[section_name, 'l_min_vec']
            upper = block[section_name, 'l_max_vec']
            angle = 0.5 * (np.log10(lower) + np.log10(upper))
            angle = 10**angle
        else:
            angle = block[section_name, angle_name]

        ## Is it GGL? If yes, the indices i & j run differently from the others.
        if isGGL:
            for i in range(nbTomo_max):
                if not block.has_value(section_name, 'bin_%d_%d' % (i+1, 1)):
                    #print('  %s stops at lens bin %d' % (section_name, i))
                    break

                for j in range(nbTomo_max):
                    value_name  = 'bin_%d_%d' % (i+1, j+1)
                    try:
                        value = block[section_name, value_name]
                        sBuilder.addTomo(i, j, angle, value)
                    except:
                        break
        else:
            for i in range(nbTomo_max):
                if not block.has_value(section_name, 'bin_%d_%d' % (i+1, i+1)):
                    #print('  %s stops at bin %d' % (section_name, i))
                    break

                for j in range(i, nbTomo_max):
                    value_name  = 'bin_%d_%d' % (j+1, i+1)
                    try:
                        value = block[section_name, value_name]
                        sBuilder.addTomo(i, j, angle, value)
                    except:
                        break

        ## These are some necessary components for generating the mean vector (spectrum).
        line  = labConv.kernelTypeDict[extension_name]
        ker1  = line[0]
        ker2  = line[1]
        type1 = line[2]
        type2 = line[3]
        unit  = line[4]

        spec  = sBuilder.makeSpectrum(extension_name, (type1, type2), unit, kernels=(ker1, ker2))
        spectra.append(spec)

    if config['onepoint_extension_name'] in config['use_stats']:
        if not block.has_section(config['onepoint_section_name']):
            raise AssertionError('You specified \"{}\" in use_stats. When I was computing for \"{}\", I tried to\n\
                grab theory values from \"{}\" but could not find anything.'.format(config['use_stats'], config['onepoint_extension_name'], config['onepoint_section_name']))
        else:
            onept_theory = [twopoint.OnePointMeasurement.from_block(block, config['onepoint_section_name'], output_name='1pt')]
    else:
        onept_theory = None

    ## Make. Ta da!
    TP_theory = wtp.TwoPointWrapper.from_spectra(spectra, nobs=onept_theory, kernels=None, covmat_info=None)

    ## Do scale cuts to theory
    scArgs = config['scale_cuts_arguments']
    TP_theory.cutScales(cutCross=scArgs[0], statsTag_tomoInd_tomoInd_list=scArgs[1], statsTag_binIndList_dict=scArgs[2])
    TP_theory.keepScales(statsTag_tomoInd1_tomoInd2__angMin_angMax_dict=scArgs[3], statsTag__angMin_angMax_dict=scArgs[4])
    #print('  Did scale cuts to theory')

    ## Extract the vector & put in block as output_section_name
    block[output_section_name, 'theory'] = TP_theory.makeMeanVector()

    if config["simulate"]:
        mu = block[output_section_name, 'theory']
        cov = block[output_section_name, 'covariance']
        if config["simulate_with_noise"]:
            if(config["number_of_simulations"]>1):
                for n in range(config["number_of_simulations"]):
                    s = np.random.multivariate_normal(mu, cov)
                    config["TP_data"].replaceMeanVector(s)
                    if config["mock_filename"] != "":
                        config["TP_data"].to_fits(config["mock_filename"]+str(n+1)+".fits", overwrite=True)
            else:
                s = np.random.multivariate_normal(mu, cov)
                config["TP_data"].replaceMeanVector(s)
                if config["mock_filename"] != "":
                    config["TP_data"].to_fits(config["mock_filename"]+".fits", overwrite=True)
        else:
            s = mu
            config["TP_data"].replaceMeanVector(s)
            if config["mock_filename"] != "":
                config["TP_data"].to_fits(config["mock_filename"]+".fits", overwrite=True)

    return 0



# This module computes the average luminosity scaling of the intrinsic alignment
# following the formalism presented in Joachimi et al. 2011b
#
#                            A(L) = (L/L_0)^beta
#
# If the luminosity distribution of the sample is not narrow, we need to average
# the contribution from L^beta of the entire sample, i.e.
#
#            <(L/L_0)^beta> = (1/L_0^beta) \int L^beta p(L) dL
#
# where p(L) is the pdf of L.
# We also allow for a double power law, i.e.
#                            A(L<L0) = (L/L_0)^beta_low
#                            A(L>L0) = (L/L_0)^beta

# -------------------------------------------------------------------------------- #
# IMPORTANT: here we assume luminosities to be in units of L_sun/h2
# -------------------------------------------------------------------------------- #

import numpy as np
from astropy.io import fits
from cosmosis.datablock import option_section
from scipy.integrate import simpson


def mean_l_l0_to_beta(xlum, pdf, l0, beta):
    return simpson(pdf * (xlum / l0) ** beta, xlum)

def broken_powerlaw(xlum, pdf, gamma_2h_lum, l0, beta, beta_low):
    alignment_ampl = np.where(xlum > l0, gamma_2h_lum * (xlum / l0) ** beta,
                              gamma_2h_lum * (xlum / l0) ** beta_low)
    return simpson(pdf * alignment_ampl, xlum)

def compute_luminosity_pdf(z_loglum_file, zmin, zmax, nz, nlbins):
    galfile = fits.open(z_loglum_file)[1].data
    z_gal = np.array(galfile['z'])
    loglum_gal = np.array(galfile['loglum'])

    z_bins = np.linspace(zmin, zmax, nz)
    dz = 0.5 * (z_bins[1] - z_bins[0])
    z_edges = np.append(z_bins - dz, z_bins[-1] + dz)

    bincen = np.zeros([nz, nlbins])
    pdf = np.zeros([nz, nlbins])

    for i in range(nz):
        mask_z = (z_gal >= z_edges[i]) & (z_gal < z_edges[i + 1])
        loglum_bin = loglum_gal[mask_z]
        if loglum_bin.size:
            lum = 10.0 ** loglum_bin
            pdf_tmp, _lum_bins = np.histogram(lum, bins=nlbins, density=True)
            _dbin = (_lum_bins[-1] - _lum_bins[0]) / (1.0 * nlbins)
            bincen[i] = _lum_bins[:-1] + 0.5 * _dbin
            pdf[i] = pdf_tmp

    return bincen, pdf

def setup(options):
    central_ia_depends_on = options[option_section, 'central_IA_depends_on']
    satellite_ia_depends_on = options[option_section, 'satellite_IA_depends_on']

    if central_ia_depends_on not in ['constant', 'luminosity', 'halo_mass']:
        raise ValueError('Choose one of the following options for central_IA_depends_on: '
                         'constant, luminosity, halo_mass')

    if satellite_ia_depends_on not in ['constant', 'luminosity', 'halo_mass']:
        raise ValueError('Choose one of the following options for satellite_IA_depends_on: '
                         'constant, luminosity, halo_mass')

    zmin = options[option_section, 'zmin']
    zmax = options[option_section, 'zmax']
    nz = options[option_section, 'nz']

    if central_ia_depends_on == 'luminosity':
        nlbins = 10000
        z_loglum_file_centrals = options[option_section, 'z_loglum_file_centrals']
        lum_centrals, lum_pdf_z_centrals = compute_luminosity_pdf(
            z_loglum_file_centrals, zmin, zmax, nz, nlbins)
    else:
        nlbins = 100000
        lum_centrals = np.ones([nz, 100000])
        lum_pdf_z_centrals = np.ones([nz, 100000])

    if satellite_ia_depends_on == 'luminosity':
        nlbins = 10000
        z_loglum_file_satellites = options[option_section, 'z_loglum_file_satellites']
        lum_satellites, lum_pdf_z_satellites = compute_luminosity_pdf(
            z_loglum_file_satellites, zmin, zmax, nz, nlbins)
    else:
        nlbins = 100000
        lum_satellites = np.ones([nz, nlbins])
        lum_pdf_z_satellites = np.ones([nz, nlbins])

    name = options.get_string(option_section, 'output_suffix', default='').lower()
    suffix = f'_{name}' if name else ''

    return lum_centrals, lum_pdf_z_centrals, lum_satellites, lum_pdf_z_satellites, nz, 10000, central_ia_depends_on, satellite_ia_depends_on, suffix

def execute(block, config):
    lum_centrals, lum_pdf_z_centrals, lum_satellites, lum_pdf_z_satellites, nz, nlum, central_ia_depends_on, satellite_ia_depends_on, suffix = config

    # TODO: Check if all the options work as intended!
    # TODO: I think we should re-write this bit so the type of dependence doesn't have to be defined
    # Include default values if parameters are not defined in the value file instead.

    # First the Centrals
    # All options require the central 2-halo amplitude to be defined
    gamma_2h = block[f'intrinsic_alignment_parameters{suffix}', 'gamma_2h_amplitude']

    if central_ia_depends_on == 'constant':
        block.put_double_array_1d(f'ia_large_scale_alignment{suffix}', 'alignment_gi',
                                  gamma_2h * np.ones(nz))

    elif central_ia_depends_on == 'luminosity':
        # Check that the user knows what they're doing:
        if not block.has_value(f'intrinsic_alignment_parameters{suffix}', 'L_pivot'):
            raise ValueError('You have chosen central luminosity scaling without providing a '
                             'pivot luminosity parameter. Include L_pivot. "\n ')
        lpiv = block[f'intrinsic_alignment_parameters{suffix}', 'L_pivot']
        beta = block[f'intrinsic_alignment_parameters{suffix}', 'beta']

        if block.has_value(f'intrinsic_alignment_parameters{suffix}', 'beta_two'):
            beta_two = block[f'intrinsic_alignment_parameters{suffix}', 'beta_two']
            mean_lscaling = np.array([broken_powerlaw(lum_centrals[i], lum_pdf_z_centrals[i],
                                                      gamma_2h, 10.0**lpiv, beta, beta_two)
                                      for i in range(nz)])
        else:
            mean_lscaling = gamma_2h * mean_l_l0_to_beta(lum_centrals, lum_pdf_z_centrals,
                                                         10.0**lpiv, beta)

        block.put_double_array_1d(f'ia_large_scale_alignment{suffix}', 'alignment_gi',
                                  mean_lscaling)

    elif central_ia_depends_on == 'halo_mass':
        # Check that the user knows what they're doing:
        if not block.has_value(f'intrinsic_alignment_parameters{suffix}', 'M_pivot'):
            raise ValueError('You have chosen central halo-mass scaling without providing a '
                             'pivot mass parameter. Include M_pivot. "\n ')
        mpiv = block.get_double(f'intrinsic_alignment_parameters{suffix}', 'M_pivot')
        beta = block.get_double(f'intrinsic_alignment_parameters{suffix}', 'beta')

        if block.has_value(f'intrinsic_alignment_parameters{suffix}', 'beta_two'):
            raise ValueError('A double power law model for the halo mass dependence of centrals '
                             'has not been implemented.')
        # Technically just repacking the variables, but this is the easiest way to accommodate
        # backwards compatibility and clean pk_lib.py module
        block.put_double_array_1d(f'ia_large_scale_alignment{suffix}', 'alignment_gi',
                                  gamma_2h * np.ones(nz))
        block.put_double(f'ia_large_scale_alignment{suffix}', 'M_pivot', 10.0**mpiv)
        block.put_double(f'ia_large_scale_alignment{suffix}', 'beta', beta)

    # Add instance information to block
    block.put_string(f'ia_large_scale_alignment{suffix}', 'instance', central_ia_depends_on)

    # Second the Satellites
    # All options require the satellite 1-halo amplitude to be defined
    gamma_1h = block[f'intrinsic_alignment_parameters{suffix}', 'gamma_1h_amplitude']

    if satellite_ia_depends_on == 'constant':
        block.put_double_array_1d(f'ia_small_scale_alignment{suffix}', 'alignment_1h',
                                  gamma_1h * np.ones(nz))

    elif satellite_ia_depends_on == 'luminosity':
        lpiv = block[f'intrinsic_alignment_parameters{suffix}', 'L_pivot']
        beta_sat = block[f'intrinsic_alignment_parameters{suffix}', 'beta_sat']
        mean_lscaling = mean_l_l0_to_beta(lum_satellites, lum_pdf_z_satellites, 10.0**lpiv,
                                          beta_sat)
        block.put_double_array_1d(f'ia_small_scale_alignment{suffix}', 'alignment_1h',
                                  gamma_1h * mean_lscaling)

    elif satellite_ia_depends_on == 'halo_mass':
        mpiv = block.get_double(f'intrinsic_alignment_parameters{suffix}', 'M_pivot')
        beta_sat = block.get_double(f'intrinsic_alignment_parameters{suffix}', 'beta_sat')
        # Technically just repacking the variables, but this is the easiest way to accommodate
        # backwards compatibility and clean pk_lib.py module
        block.put_double_array_1d(f'ia_small_scale_alignment{suffix}', 'alignment_1h',
                                  gamma_1h * np.ones(nz))
        block.put_double(f'ia_small_scale_alignment{suffix}', 'M_pivot', 10.0**mpiv)
        block.put_double(f'ia_small_scale_alignment{suffix}', 'beta_sat', beta_sat)

    # Add instance information to block
    block.put_string(f'ia_small_scale_alignment{suffix}', 'instance', satellite_ia_depends_on)

    return 0

def cleanup(config):
    pass



# Library of the power spectrum module
import numexpr as ne
import numpy as np
import warnings
from scipy.integrate import quad, simpson, trapezoid
from scipy.interpolate import RegularGridInterpolator, UnivariateSpline, interp1d
from scipy.ndimage import gaussian_filter1d

# from scipy.special import erf
from scipy.optimize import curve_fit


def low_k_truncation(k_vec, k_trunc):
    """
    Beta_nl low-k truncation
    """
    return 1.0 / (1.0 + np.exp(-(10.0 * (np.log10(k_vec) - np.log10(k_trunc)))))

def high_k_truncation(k_vec, k_trunc):
    """
    Beta_nl high-k truncation
    """
    return 1.0 / (1.0 + np.exp(10.0 * (np.log10(k_vec) - np.log10(k_trunc))))

def minimum_halo_mass(emu):
    """
    Minimum halo mass for the set of cosmological parameters [Msun/h]
    """
    np_min = 200.0 # Minimum number of halo particles
    npart = 2048.0 # Cube root of number of simulation particles
    Lbox_HR = 1000.0 # Box size for high-resolution simulations [Mpc/h]
    Lbox_LR = 2000.0 # Box size for low-resolution simulations [Mpc/h]

    Om_m = emu.cosmo.get_Omega0()
    rhom = 2.77536627e11 * Om_m

    Mbox_HR = rhom * Lbox_HR**3.0
    mmin = Mbox_HR * np_min / npart**3.0

    vmin = Lbox_HR**3.0 * np_min / npart**3.0
    rmin = ((3.0 * vmin) / (4.0 * np.pi))**(1.0 / 3.0)

    return mmin, 2.0 * np.pi / rmin

def rvir(emu, mass):
    Om_m = emu.cosmo.get_Omega0()
    rhom = 2.77536627e11 * Om_m
    return ((3.0 * mass) / (4.0 * np.pi * 200 * rhom))**(1.0 / 3.0)

def hl_envelopes_idx(data, dmin=1, dmax=1):
    """
    Extract high and low envelope indices from a 1D data signal.

    Parameters:
    data (1d-array): Data signal from which to extract high and low envelopes.
    dmin (int): Size of chunks for local minima, use this if the size of the input signal is too big.
    dmax (int): Size of chunks for local maxima, use this if the size of the input signal is too big.

    Returns:
    lmin, lmax (tuple of arrays): Indices of high and low envelopes of the input signal.
    """
    # Find local minima indices
    lmin = (np.diff(np.sign(np.diff(data))) > 0).nonzero()[0] + 1
    # Find local maxima indices
    lmax = (np.diff(np.sign(np.diff(data))) < 0).nonzero()[0] + 1

    # Global min of dmin-chunks of local minima
    lmin = lmin[[i + np.argmin(data[lmin[i:i + dmin]]) for i in range(0, len(lmin), dmin)]]
    # Global max of dmax-chunks of local maxima
    lmax = lmax[[i + np.argmax(data[lmax[i:i + dmax]]) for i in range(0, len(lmax), dmax)]]

    return lmin, lmax

def compute_bnl_darkquest(z, log10M1, log10M2, k, emulator, block, kmax):
    M1 = 10.0**log10M1
    M2 = 10.0**log10M2

    # Large 'linear' scale for linear halo bias [h/Mpc]
    klin = np.array([0.05])

    # Calculate beta_NL by looping over mass arrays
    beta_func = np.zeros((len(M1), len(M2), len(k)))

    # Linear power
    Pk_lin = emulator.get_pklin_from_z(k, z)
    Pk_klin = emulator.get_pklin_from_z(klin, z)

    # Calculate b01 for all M1
    b01 = np.zeros(len(M1))
    #b02 = np.zeros(len(M2))
    for iM, M0 in enumerate(M1):
        b01[iM] = np.sqrt(emulator.get_phh_mass(klin, M0, M0, z) / Pk_klin)

    for iM1, M01 in enumerate(M1):
        for iM2, M02 in enumerate(M2):
            if iM2 < iM1:
                # Use symmetry to not double calculate
                beta_func[iM1, iM2, :] = beta_func[iM2, iM1, :]
            else:
                # Linear halo bias
                b1 = b01[iM1]
                b2 = b01[iM2]

                # Halo-halo power spectrum
                Pk_hh = emulator.get_phh_mass(k, M01, M02, z)

                #rmax = max(rvir(emulator, M01), rvir(emulator, M02))
                #kmax = 2.0*np.pi/rmax

                # Create beta_NL
                shot_noise = lambda x, a: a
                popt, popc = curve_fit(shot_noise, k[(k > 100) & (k < 200)], Pk_hh[(k > 100) & (k < 200)])
                Pk_hh = Pk_hh - np.ones_like(k) * shot_noise(k, *popt)

                beta_func[iM1, iM2, :] = Pk_hh / (b1 * b2 * Pk_lin) - 1.0

                Pk_hh0 = emulator.get_phh_mass(klin, M01, M02, z)
                Pk_hh0 = Pk_hh0 - np.ones_like(klin)*shot_noise(klin, *popt)
                db = Pk_hh0 / (b1 * b2 * Pk_klin) - 1.0

                lmin, lmax = hl_envelopes_idx(np.abs(beta_func[iM1, iM2, :]+1.0))
                beta_func_interp = interp1d(k[lmax], np.abs(beta_func[iM1, iM2, lmax]+1.0), kind='quadratic', bounds_error=False, fill_value='extrapolate')
                beta_func[iM1, iM2, :] = (beta_func_interp(k) - 1.0)# * low_k_truncation(k, klin)
                db = (beta_func_interp(klin) - 1.0)


                #beta_func[iM1, iM2, :] = ((beta_func[iM1, iM2, :] + 1.0) * high_k_truncation(k, 30.0)/(db + 1.0) - 1.0) * low_k_truncation(k, klin)
                #beta_func[iM1, iM2, :] = ((beta_func[iM1, iM2, :] + 1.0)/(db + 1.0) - 1.0) #* low_k_truncation(k, klin) * high_k_truncation(k, 30.0)#/(1.0+z))
                beta_func[iM1, iM2, :] = (beta_func[iM1, iM2, :] - db) * low_k_truncation(k, klin) * high_k_truncation(k, 3.0*kmax)

    return beta_func

def create_bnl_interpolation_function(emulator, interpolation, z, block):
    lenM = 5
    lenk = 1000
    zc = z.copy()

    Mmin, kmax = minimum_halo_mass(emulator)
    M_up = np.log10(10.0**14.0)
    #M_lo = np.log10((10.0**12.0))
    M_lo = np.log10(Mmin)

    M = np.logspace(M_lo, M_up, lenM)
    k = np.logspace(-3.0, np.log10(200), lenk)
    beta_nl_interp_i = np.empty(len(z), dtype=object)
    beta_func = compute_bnl_darkquest(0.01, np.log10(M), np.log10(M), k, emulator, block, kmax)
    beta_nl_interp_i = RegularGridInterpolator([np.log10(M), np.log10(M), np.log10(k)], beta_func, fill_value=None, bounds_error=False, method='nearest')
    """
    for i,zi in enumerate(zc):
        #M = np.logspace(M_lo, M_up - 3.0*np.log10(1+zi), lenM)
        #beta_func = compute_bnl_darkquest(zi, np.log10(M), np.log10(M), k, emulator, block, kmax)
        beta_nl_interp_i[i] = RegularGridInterpolator([np.log10(M), np.log10(M), np.log10(k)],
                                                      beta_func, fill_value=None, bounds_error=False, method='nearest')
    """
    return beta_nl_interp_i



import numpy as np
import warnings
from astropy.cosmology import Planck15
from scipy.integrate import quad, simpson, solve_ivp
from scipy.interpolate import interp1d
from scipy.optimize import root_scalar

from hmf.cosmology.cosmo import astropy_to_colossus
from hmf.halos.mass_definitions import SphericalOverdensity

warnings.filterwarnings('ignore', category=UserWarning, module='colossus')

# TODO: unused
def concentration_halomod(cosmo, mass, z, model, mdef, overdensity, mf, delta_c):
    """
    Calculate concentration given halo mass using the halomod model provided in config.
    """
    mdef = getattr(md, mdef)() if mdef in ['SOVirial'] else getattr(md, mdef)(overdensity=overdensity)
    cm = getattr(conc_func, model)(cosmo=mf, filter0=mf.filter, delta_c=delta_c, mdef=mdef)
    return cm.cm(mass, z)

def get_halo_collapse_redshifts(M, z, dc, g, cosmo, mf):
    """
    Calculate halo collapse redshifts according to the Bullock et al. (2001) prescription.
    """
    gamma = 0.01
    a = cosmo.scale_factor(z)
    zf = np.zeros_like(M)
    for iM, _M in enumerate(M):
        Mc = gamma * _M
        Rc = mf.filter.mass_to_radius(Mc, mf.mean_density0)
        sigma = mf.normalised_filter.sigma(Rc)
        fac = g(a) * dc / sigma
        if fac >= g(a):
            af = a  # These haloes formed 'in the future'
        else:
            af_root = lambda af: g(af) - fac
            af = root_scalar(af_root, bracket=(1e-3, 1.)).root
        zf[iM] = -1.0 + 1.0 / af
    return zf

def acceleration_parameter(cosmo, z):
    """
    Calculate the acceleration parameter.
    """
    return -0.5 * (cosmo.Om(z) + (1.0 + 3.0 * cosmo.w(z)) * cosmo.Ode(z))

def _w(a, w0, wa):
    """
    Dark energy equation of state for w0, wa models.
    """
    return w0 + (1. - a) * wa

def _X_w(a, w0, wa):
    """
    Cosmological dark energy density for w0, wa models.
    """
    return a ** (-3. * (1. + w0 + wa)) * np.exp(-3. * wa * (1. - a))

def _Omega_m(a, Om, Ode, Ok, w0, wa):
    """
    Evolution of Omega_m with scale-factor ignoring radiation.
    """
    return Om * a**-3 / _Hubble2(a, Om, Ode, Ok, w0, wa)

def _Hubble2(a, Om, Ode, Ok, w0, wa):
    """
    Squared Hubble parameter ignoring radiation.
    Massive neutrinos are counted as 'matter'.
    """
    H2 = Om * a**-3 + Ode * _X_w(a, w0, wa) + Ok * a**-2
    return H2

def _AH(a, Om, Ode, w0, wa):
    """
    Acceleration parameter ignoring radiation.
    Massive neutrinos are counted as 'matter'.
    """
    AH = -0.5 * (Om * a**-3 + (1. + 3. * _w(a, w0, wa)) * Ode * _X_w(a, w0, wa))
    return AH

def get_growth_interpolator(cosmo):
    """
    Solve the linear growth ODE and returns an interpolating function for the solution.
    TODO: w dependence for initial conditions; f here is correct for w=0 only.
    TODO: Could use d_init = a(1+(w-1)/(w(6w-5))*(Om_w/Om_m)*a**-3w) at early times with w = w(a<<1).
    """
    a_init = 1e-4
    Om = cosmo.Om0 + cosmo.Onu0
    Ode = 1.0 - Om
    Ok = cosmo.Ok0
    w0 = getattr(cosmo, 'w0', -1.0)
    wa = getattr(cosmo, 'wa', 0.0)
    na = 129  # Number of scale factors used to construct interpolator
    a = np.linspace(a_init, 1., na)
    f = 1. - _Omega_m(a_init, Om, Ode, Ok, w0, wa)  # Early mass density
    d_init = a_init**(1. - 3. * f / 5.)  # Initial condition (~ a_init; but f factor accounts for EDE-ish)
    v_init = (1. - 3. * f / 5.) * a_init**(-3. * f / 5.)  # Initial condition (~ 1; but f factor accounts for EDE-ish)
    y0 = (d_init, v_init)

    def fun(a, y):
        d, v = y[0], y[1]
        dxda = v
        fv = -(2. + _AH(a, Om, Ode, w0, wa) / _Hubble2(a, Om, Ode, Ok, w0, wa)) * v / a
        fd = 1.5 * _Omega_m(a, Om, Ode, Ok, w0, wa) * d / a**2
        dvda = fv + fd
        return dxda, dvda

    g = solve_ivp(fun, (a[0], a[-1]), y0, t_eval=a).y[0]
    return interp1d(a, g, kind='cubic', assume_sorted=True)

#def get_growth_interpolator(cosmo):
#    """
#    Solve the linear growth ODE and returns an interpolating function for the solution
#    LCDM = True forces w = -1 and imposes flatness by modifying the dark-energy density
#    TODO: w dependence for initial conditions; f here is correct for w=0 only
#    TODO: Could use d_init = a(1+(w-1)/(w(6w-5))*(Om_w/Om_m)*a**-3w) at early times with w = w(a<<1)
#    """
#    a_init = 1e-3
#    z_init = (1.0/a_init) - 1.0
#    print(z_init)
#    #z_init = 500.0
#    #a_init = 1.0/(1.0+z_init)
#
#    na = 129 # Number of scale factors used to construct interpolator
#    a = np.linspace(a_init, 1.0, na)
#    f = 1.0 - cosmo.Om(z_init) # Early mass density
#    d_init = a_init**(1.0 - ((3.0/5.0) * f))            # Initial condition (~ a_init; but f factor accounts for EDE-ish)
#    v_init = (1.0 - ((3.0/5.0) * f))*a_init**(-((3.0/5.0) * f)) # Initial condition (~ 1; but f factor accounts for EDE-ish)
#
#    y0 = (d_init, v_init)
#    def fun(ax, y):
#        d, v = y[0], y[1]
#        dxda = v
#        zx = (1.0/ax) - 1.0
#        fv = -(2.0 + acceleration_parameter(cosmo, zx)*cosmo.inv_efunc(zx)**2.0)*v/ax
#        fd = 1.5*cosmo.Om(zx)*d/ax**2
#        dvda = fv+fd
#        return dxda, dvda
#
#    #g = solve_ivp(fun, (a[0], a[-1]), y0, t_eval=a, atol=1e-8, rtol=1e-8, vectorized=True).y[0]
#    g = solve_ivp(fun, (a[0], a[-1]), y0, t_eval=a).y[0]
#    g_interp = interp1d(a, g, kind='linear', assume_sorted=True)
#    return g_interp

# Mead corrections: See appendix A of 2009.01858
def get_accumulated_growth(a, g):
    """
    Calculates the accumulated growth at scale factor 'a'.
    """
    a_init = 1e-4

    # Eq A5 of Mead et al. 2021 (2009.01858).
    # We approximate the integral as g(a_init) for 0 to a_init<<0.
    missing = g(a_init)
    G, _ = quad(lambda a: g(a) / a, a_init, a) + missing
    return G

def f_Mead(x, y, p0, p1, p2, p3):
    # eq A3 of 2009.01858
    return p0 + p1 * (1.0 - x) + p2 * (1.0 - x)**2.0 + p3 * (1.0 - y)

def dc_Mead(a, Om, f_nu, g, G):
    """
    delta_c fitting function from Mead et al. 2021 (2009.01858).
    All input parameters should be evaluated as functions of a/z.
    """
    # See Table A.1 of 2009.01858 for naming convention
    p1 = [-0.0069, -0.0208, 0.0312, 0.0021]
    p2 = [0.0001, -0.0647, -0.0417, 0.0646]
    a1, a2 = 1, 0
    # Linear collapse threshold
    # Eq A1 of 2009.01858
    dc_Mead = 1.0 + f_Mead(g / a, G / a, *p1) * np.log10(Om) ** a1 + f_Mead(g / a, G / a, *p2) * np.log10(Om) ** a2
    # delta_c = ~1.686' EdS linear collapse threshold
    dc0 = (3.0 / 20.0) * (12.0 * np.pi)**(2.0 / 3.0)
    return dc_Mead * dc0 * (1.0 - 0.041 * f_nu)

def Dv_Mead(a, Om, f_nu, g, G):
    """
    Delta_v fitting function from Mead et al. 2021 (2009.01858), eq A.2.
    All input parameters should be evaluated as functions of a/z.
    """
    # See Table A.1 of 2009.01858 for naming convention
    p3 = [-0.79, -10.17, 2.51, 6.51]
    p4 = [-1.89, 0.38, 18.8, -15.87]
    a3, a4 = 1, 2

    # Halo virial overdensity
    # Eq A2 of 2009.01858
    Dv_Mead = 1.0 + f_Mead(g / a, G / a, *p3) * np.log10(Om) ** a3 + f_Mead(g / a, G / a, *p4) * np.log10(Om) ** a4
    Dv0 = 18.0 * np.pi**2.0  # Delta_v = ~178, EdS halo virial overdensity
    return Dv_Mead * Dv0 * (1.0 + 0.763 * f_nu)

def Tk_cold_ratio(k, g, ommh2, h, f_nu, N_nu, T_CMB=2.7255):
    """
    Ratio of cold to matter transfer function from Eisenstein & Hu (1999).
    This can be used to get the cold-matter spectrum approximately from the matter spectrum.
    Captures the scale-dependent growth with neutrino free-streaming scale.
    """
    if f_nu == 0.:  # Fix to unity if there are no neutrinos
        return 1.

    pcb = (5. - np.sqrt(1. + 24. * (1. - f_nu))) / 4.  # Growth exponent for unclustered neutrinos completely
    BigT = T_CMB / 2.7  # Big Theta for temperature
    zeq = 2.5e4 * ommh2 * BigT**(-4)  # Matter-radiation equality redshift
    D = (1. + zeq) * g  # Growth normalized such that D=(1.+z_eq)/(1+z) at early times
    q = k * h * BigT**2 / ommh2  # Wave number relative to the horizon scale at equality (equation 5)
    yfs = 17.2 * f_nu * (1. + 0.488 * f_nu**(-7. / 6.)) * (N_nu * q / f_nu)**2  # Free streaming scale (equation 14)
    Dcb = (1. + (D / (1. + yfs))**0.7)**(pcb / 0.7)  # Cold growth function
    Dcbnu = ((1. - f_nu)**(0.7 / pcb) + (D / (1. + yfs))**0.7)**(pcb / 0.7)  # Cold and neutrino growth function
    return Dcb / Dcbnu  # Finally, the ratio

def sigmaR_cc(power, k, r):
    rk = np.outer(r, k)
    dlnk = np.log(k[1] / k[0])

    k_space = (3 / rk**3) * (np.sin(rk) - rk * np.cos(rk))
    # we multiply by k because our steps are in logk.
    rest = power * k ** 3
    integ = rest * k_space ** 2
    sigma = (0.5 / np.pi**2) * simpson(integ, dx=dlnk, axis=-1)
    return np.sqrt(sigma)

# To be maybe incorporated in hmf
class SOVirial_Mead(SphericalOverdensity):
    """
    SOVirial overdensity definition from Mead et al. (2021).
    """
    _defaults = {"overdensity": 200}

    def halo_density(self, z=0, cosmo=Planck15):
        """The density of haloes under this definition."""
        return self.params["overdensity"] * self.mean_density(z, cosmo)

    @property
    def colossus_name(self):
        return "200c"

    def __str__(self):
        """Describe the halo definition in standard notation."""
        return "SOVirial"


# def concentration_colossus(block, cosmo, mass, z, model, mdef, overdensity):
#     """
#     calculates concentration given halo mass, using the halomod model provided in config
#     furthermore it converts to halomod instance to be used with the halomodel,
#     consistenly with halo mass function and halo bias function
#     """

#     with warnings.catch_warnings():
#         warnings.filterwarnings('ignore', category=UserWarning)
#         # This dissables the warning from colossus that is just telling us what we know
#         # colossus warns us about massive neutrinos, but that is also ok
#         # as we do not use cosmology from it, but it requires it to setup the instance!
#         this_cosmo = colossus_cosmology.fromAstropy(astropy_cosmo=cosmo, cosmo_name='custom',
#                      sigma8=block[cosmo_params, 'sigma_8'], ns=block[cosmo_params, 'n_s'])


#     mdef = getattr(md, mdef)() if mdef in ['SOVirial'] else getattr(md, mdef)(overdensity=overdensity)

#     # This is the slow part: 0.4-0.5 seconds per call, called separately for each redshift.
#     # MA: Possible solution: See if we can get away with a smaller numbr of redshifts and interpolate.
#     #tic = time.perf_counter()
#     c, ms = colossus_concentration.concentration(M=mass, z=z, mdef=mdef.colossus_name, model=model,
#             range_return=True, range_warning=False)
#     #toc = time.perf_counter()
#     #print(" colossus_concentration.concentration: "+'%.4f' %(toc - tic)+ "s")
#     if len(c[c>0]) == 0:
#         c_interp = lambda x: np.ones_like(x)
#     else:
#         c_interp = interp1d(mass[c>0], c[c>0], kind='linear', bounds_error=False, fill_value=1.0)
#     return c_interp(mass)



"""
This module has two modes: response and direct.
If in direct mode it will directly calculate the power spectra.
If in response mode it will calculate the response of the halo model to different power spectra with respect to P_mm
and multiply that to an input P_mm to estimate the desired power, for example:
res_gg = P^hm_gg / P^hm_mm
res_gm = P^hm_gm / P^hm_mm
Then uses these in combination with input matter power spectra, P_mm, to create 3D power for P_gg and P_gm:
P_gg = res_gg * P_mm
P_gm = res_gm * P_mm

The following explains how P^hm_xy are calculated.
Calculates 3D power spectra using the halo model approach:
See section 2 of https://arxiv.org/pdf/2303.08752.pdf for details.

P_uv = P^2h_uv + P^1h_uv  (1)

P^1h_uv(k) = int_0^infty dM Wu(M, k) Wv(M, k) n(M)  (2)

P^2h_uv(k) = int_0^infty int_0^infty dM1 dM2 Phh(M1, M2, k) Wu(M1, k) Wv(M2, k) n(M1) n(M2)  (3)

Wx are the profile of the fields, u and v, showing how they fit into haloes.
n(M) is the halo mass function, quantifying the number of haloes of each mass, M.
Integrals are taken over halo mass.

The halo-halo power spectrum can be written as,

Phh(M1, M2, k) = b(M1) b(M2) P^lin_mm(k) (1 + beta_nl(M1, M2, k)) (4)

In the vanilla halo model, the 2-halo term is usually simplified by assuming that haloes are linearly biased with respect to matter.
This sets beta_nl to zero and effectively decouples the integrals in (3). Here we allow for both options to be calculated.
If you want the option with beta_nl, the beta_nl module has to be run before this module.

We truncate the 1-halo term so that it doesn't dominate at large scales.

Linear matter power spectrum needs to be provided as well. The halo_model_ingredients and hod modules (for everything but mm)
need to be run before this.

Current power spectra that we predict are:
mm: matter-matter
gg: galaxy-galaxy
gm: galaxy-matter

II: intrinsic-intrinsic alignments
gI: galaxy-intrinsic alignment
mI: matter-intrinsic alignment
"""

# TODO: IMPORTANT 1h term is too small compared to mead2020, see where this is coming from
# MA: It is smaller when mnu > 0 because of the fnu factor in matter profile. It might be a better approximation
# compared to simulations with neutrinos. We need to check this. I tried to compare with Euclid and Bacco emulator but
# had difficulties running them.

# NOTE: no truncation (halo exclusion problem) applied!

import numbers
import numpy as np
import pk_lib
from cosmosis.datablock import names, option_section

cosmo_params = names.cosmological_parameters



def get_string_or_none(cosmosis_block, section, name, default):
    """
    A helper function to return a number or None explicitly from config files
    or return None if no value is present.
    """
    if cosmosis_block.has_value(section, name):
        test_param = cosmosis_block.get(section, name)
        if isinstance(test_param, numbers.Number):
            param = cosmosis_block.get_double(section, name, default)
        if isinstance(test_param, str):
            str_in = cosmosis_block.get_string(section, name)
            if str_in == 'None':
                param = None
    else:
        try:
            param = cosmosis_block.get_double(section, name, default)
        except:
            param = None

    if not isinstance(param, (numbers.Number, type(None))):
        raise ValueError(f'Parameter {name} is not an instance of a number or NoneType!')

    return param

def setup(options):
    p_mm = options.get_bool(option_section, 'p_mm', default=False)
    p_gg = options.get_bool(option_section, 'p_gg', default=False)
    p_gm = options.get_bool(option_section, 'p_gm', default=False)
    p_gI = options.get_bool(option_section, 'p_gI', default=False)
    p_mI = options.get_bool(option_section, 'p_mI', default=False)
    p_II = options.get_bool(option_section, 'p_II', default=False)

    # If True, calculate the response of the halo model for the requested power spectra compared to matter power
    # multiplies this to input non-linear matter power spectra.
    response = options.get_bool(option_section, 'response', default=False)

    # If true, use the IA formalism of Fortuna et al. 2021: Truncated NLA at high k + 1-halo term
    fortuna = options.get_bool(option_section, 'fortuna', default=False)
    # If True, uses beta_nl
    bnl = options.get_bool(option_section, 'bnl', default=False)

    poisson_type = options.get_string(option_section, 'poisson_type', default='')
    point_mass = options.get_bool(option_section, 'point_mass', default=False)

    dewiggle = options.get_bool(option_section, 'dewiggle', default=False)

    # Fortuna introduces a truncation of the 1-halo term at large scales to avoid the halo exclusion problem
    # and a truncation of the NLA 2-halo term at small scales to avoid double-counting of the 1-halo term
    # The user can change these values.
    one_halo_ktrunc_ia = get_string_or_none(options, option_section, 'one_halo_ktrunc_ia', default=4.0)  # h/Mpc or None
    two_halo_ktrunc_ia = get_string_or_none(options, option_section, 'two_halo_ktrunc_ia', default=6.0)  # h/Mpc or None
    # General truncation of non-IA terms:
    one_halo_ktrunc = get_string_or_none(options, option_section, 'one_halo_ktrunc', default=0.1)  # h/Mpc or None
    two_halo_ktrunc = get_string_or_none(options, option_section, 'two_halo_ktrunc', default=2.0)  # h/Mpc or None

    # Initiate pipeline parameters
    matter = False
    galaxy = False
    alignment = False

    hod_section_name = options.get_string(option_section, 'hod_section_name')

    matter = p_mm or p_gm or p_mI
    galaxy = p_gg or p_gm or p_gI or p_mI or p_II
    alignment = p_gI or p_mI or p_II

    population_name = options.get_string(option_section, 'output_suffix', default='').lower()
    pop_name = f'_{population_name}' if population_name else ''

    # Option to set similar corrections to HMcode2020
    use_mead = options.get_string(option_section, 'use_mead2020_corrections', default='None')

    # Mapping of use_mead values to mead_correction values
    mead_correction_map = {
        'mead2020': 'nofeedback',
        'mead2020_feedback': 'feedback',
        'fit_feedback': 'fit',
        # Add more mappings here if needed
    }

    # Determine the mead_correction based on the mapping
    mead_correction = mead_correction_map.get(use_mead, None)
    if mead_correction == 'fit':
        if not options.has_value(option_section, 'hod_section_name'):
            raise ValueError('To use the fit option for feedback that links HOD derived stellar mass fraction to the baryon '
                             'feedback one needs to provide the hod section name of used hod!')

    return p_mm, p_gg, p_gm, p_gI, p_mI, p_II, response, fortuna, matter, galaxy, bnl, alignment, one_halo_ktrunc, two_halo_ktrunc, one_halo_ktrunc_ia, two_halo_ktrunc_ia, hod_section_name, mead_correction, dewiggle, point_mass, poisson_type, pop_name

def execute(block, config):
    p_mm, p_gg, p_gm, p_gI, p_mI, p_II, response, fortuna, matter, galaxy, bnl, alignment, one_halo_ktrunc, two_halo_ktrunc, one_halo_ktrunc_ia, two_halo_ktrunc_ia, hod_section_name, mead_correction, dewiggle, point_mass, poisson_type, pop_name = config

    # Load the halo mass, halo bias, mass, and redshifts from the datablock
    dn_dlnm, b_dm, mass, z_vec = pk_lib.get_halo_functions(block)

    # Reads in the Fourier transform of the normalized dark matter halo profile
    u_dm, u_sat, k_vec = pk_lib.get_normalised_profile(block, mass, z_vec)

    # Load the linear power spectrum and growth factor
    k_vec_original, plin_original = pk_lib.get_linear_power_spectrum(block, z_vec)
    plin = pk_lib.log_linear_interpolation_k(plin_original, k_vec_original, k_vec)
    growth_factor, scale_factor = pk_lib.get_growth_factor(block, z_vec, k_vec)

    if response:
        k_nl, p_nl = pk_lib.get_nonlinear_power_spectrum(block, z_vec)
        pk_mm_in = pk_lib.log_linear_interpolation_k(p_nl, k_nl, k_vec)

    # Optionally de-wiggle linear power spectrum as in Mead 2020:
    if mead_correction in ['feedback', 'nofeedback'] or dewiggle:
        plin = pk_lib.dewiggle(plin, k_vec, block)

    # AD: The following line is only used for testing, will be removed when we start running final chains!
    #k_nl, p_nl = pk_lib.get_nonlinear_power_spectrum(block, z_vec)
    #pk_mm_in = pk_lib.log_linear_interpolation_k(p_nl, k_nl, k_vec)
    #block.replace_grid('matter_power_nl_mead', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_mm_in)

    # Add the non-linear P_hh to the 2h term
    if bnl:
        # Reads beta_nl from the block
        if block.has_value('bnl', 'beta_interp'):
            beta_interp = block.get_double_array_nd('bnl', 'beta_interp')
        else:
            raise Exception("You've set bnl = True. Looked for beta_intep in bnl, but didn't find it. "
                            "Run bnl_interface.py to set this up.\n")
        if beta_interp.shape == np.array([0.0]).shape:
            raise ValueError('Non-linear halo bias module bnl is not initialized, or you have deleted it too early! '
                             'This might be because you ran bnl_interface_delete.py before this module. \n')

        integrand_12 = pk_lib.prepare_I12_integrand(b_dm, b_dm, mass, mass, dn_dlnm, dn_dlnm, beta_interp)
        integrand_21 = pk_lib.prepare_I21_integrand(b_dm, b_dm, mass, mass, dn_dlnm, dn_dlnm, beta_interp)
        integrand_22 = pk_lib.prepare_I22_integrand(b_dm, b_dm, mass, mass, dn_dlnm, dn_dlnm, beta_interp)

    # Accounts for the missing low mass haloes in the integrals for the 2h term.
    # Assumes all missing mass is in haloes of mass M_min.
    # This is calculated separately for each redshift
    # TODO: check if this is needed for the IA section
    mean_density0 = block['density', 'mean_density0'] * np.ones(len(z_vec))
    A_term = pk_lib.missing_mass_integral(mass, b_dm, dn_dlnm, mean_density0)

    # f_nu = omega_nu / omega_m with the same length as redshift
    fnu = block[cosmo_params, 'fnu'] * np.ones(len(z_vec))
    omega_c = block[cosmo_params, 'omega_c']
    omega_m = block[cosmo_params, 'omega_m']
    omega_b = block[cosmo_params, 'omega_b']

    # If matter auto or cross power spectra are set to True
    if matter:
        # 2h term integral for matter
        I_m = pk_lib.Im_term(mass, u_dm, b_dm, dn_dlnm, mean_density0, A_term)
        # Matter halo profile
        matter_profile = pk_lib.matter_profile(mass, mean_density0, u_dm, np.zeros_like(fnu))

        if mead_correction == 'feedback':
            log10T_AGN = block['halo_model_parameters', 'logT_AGN']
            matter_profile_1h_mm = pk_lib.matter_profile_with_feedback(mass, mean_density0, u_dm, z_vec, omega_c,
                                                                       omega_m, omega_b, log10T_AGN, fnu)
        elif mead_correction == 'fit':
            # Reads f_star_extended form the HOD section. Either need to use a conditional HOD to get this value or to put it in block some other way.
            fstar_mm = pk_lib.load_fstar_mm(block, hod_section_name, z_vec, mass)
            mb = 10.0**block['halo_model_parameters', 'm_b']
            matter_profile_1h_mm = pk_lib.matter_profile_with_feedback_stellar_fraction_from_obs(mass, mean_density0, u_dm, z_vec, mb,
                                                                                                 fstar_mm, omega_c, omega_m, omega_b, fnu)
        else:
            matter_profile_1h_mm = pk_lib.matter_profile(mass, mean_density0, u_dm, fnu)

        if bnl:
            I_NL_mm = pk_lib.I_NL(mass, mass, matter_profile, matter_profile, b_dm, b_dm,
                                  dn_dlnm, dn_dlnm, A_term, mean_density0, beta_interp, integrand_12, integrand_21, integrand_22)

        if p_mm or response:
            if bnl:
                pk_mm_1h, pk_mm_2h, pk_mm = pk_lib.compute_p_mm_bnl(k_vec, plin, mass, dn_dlnm,
                                                                   matter_profile_1h_mm, I_m, I_NL_mm,
                                                                   one_halo_ktrunc)
            else:
                if mead_correction in ['feedback', 'nofeedback']:
                    sigma8_z = block['hmf', 'sigma8_z']
                    neff = block['hmf', 'neff']
                    pk_mm_1h, pk_mm_2h, pk_mm = pk_lib.compute_p_mm_mead(k_vec, plin, mass,
                                                                         dn_dlnm, matter_profile_1h_mm,
                                                                         sigma8_z, neff)
                else:
                    pk_mm_1h, pk_mm_2h, pk_mm = pk_lib.compute_p_mm(k_vec, plin, mass,
                                                                    dn_dlnm, matter_profile_1h_mm, I_m,
                                                                    one_halo_ktrunc, two_halo_ktrunc)
            if response:
                # Here we save the computed Pmm to datablock as matter_power_hm,
                # but not replacing the Pnl with it, as in the response
                # method, the Pnl stays the same as one from CAMB
                block.put_grid('matter_power_hm', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_mm_1h)
                block.put_grid('matter_power_hm', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_mm_2h)
                block.put_grid('matter_power_hm', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_mm)
            else:
                block.put_grid('matter_power_nl', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_mm_1h)
                block.put_grid('matter_power_nl', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_mm_2h)
                block.put_grid('matter_power_nl', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_mm)

    if galaxy or alignment:
        hod_bins = block[hod_section_name, 'nbins']

        # Check number of observable-redshift bins and read in the input for the HOD of each bin
        for nb in range(hod_bins):
            suffix = f'{pop_name}_{nb+1}' if hod_bins != 1 else f'{pop_name}'
            suffix_hod = f'_{nb+1}' if hod_bins != 1 else ''

            N_cen, N_sat, numdencen, numdensat, f_cen, f_sat, mass_avg, fstar = pk_lib.load_hods(block,
                                                                                                 hod_section_name, suffix_hod, z_vec, mass)

            if galaxy:
                # Preparing the 1-halo term
                # Computes the profiles for centrals and satellites.
                # These are the W_u(M,k) functions in Asgari, Mead, Heymans 2023: 2303.08752
                # We assume that the centrals are in the centre of the halo,
                # and set their normalized profile to 1 everywhere.

                profile_c = pk_lib.galaxy_profile(N_cen, numdencen, f_cen, np.ones_like(u_sat))
                profile_s = pk_lib.galaxy_profile(N_sat, numdensat, f_sat, u_sat)

                # Calculate the 2-halo integrals for centrals and satellites
                I_c = pk_lib.Ig_term(mass, profile_c, b_dm, dn_dlnm)
                I_s = pk_lib.Ig_term(mass, profile_s, b_dm, dn_dlnm)

                if mead_correction == 'fit' or point_mass:
                    # Include point mass and gas contribution to the GGL power spectrum, defined from HOD
                    # Maybe extend to input the mass per bin!
                    mb = 10.0**block['halo_model_parameters', 'm_b']
                    matter_profile_1h = pk_lib.matter_profile_with_feedback_stellar_fraction_from_obs(mass, mean_density0,
                                                                                                    u_dm, z_vec, mb, fstar,
                                                                                                    omega_c, omega_m, omega_b, fnu)
                else:
                    matter_profile_1h = matter_profile_1h_mm.copy()

                if bnl:
                    if p_gg:
                        I_NL_gg = pk_lib.I_NL(mass, mass, profile_c + profile_s, profile_c + profile_s, b_dm, b_dm,
                                              dn_dlnm, dn_dlnm, A_term, mean_density0, beta_interp, integrand_12, integrand_21, integrand_22)
                    if p_gm:
                        I_NL_gm = pk_lib.I_NL(mass, mass, profile_c + profile_s, matter_profile, b_dm, b_dm,
                                              dn_dlnm, dn_dlnm, A_term, mean_density0, beta_interp, integrand_12, integrand_21, integrand_22)

            if alignment:
                # Load the 2h (effective) amplitude of the alignment signal from the data block.
                alignment_gi = block[f'ia_large_scale_alignment{pop_name}', 'alignment_gi']
                alignment_amplitude_2h, alignment_amplitude_2h_II, C1 = pk_lib.compute_two_halo_alignment(alignment_gi, growth_factor, mean_density0)
                wkm = pk_lib.get_satellite_alignment(block, k_vec, mass, z_vec, pop_name)
                # Preparing the central and satellite terms
                if block[f'ia_small_scale_alignment{pop_name}', 'instance'] == 'halo_mass':
                    beta_sat = block[f'ia_small_scale_alignment{pop_name}', 'beta_sat']
                    M_pivot = block[f'ia_small_scale_alignment{pop_name}', 'M_pivot']
                    s_align_profile = pk_lib.satellite_alignment_profile(N_sat, numdensat, f_sat, wkm,
                                                                         beta_sat, M_pivot, mass_avg)
                else:
                    s_align_profile = pk_lib.satellite_alignment_profile(N_sat, numdensat, f_sat, wkm)
                if block[f'ia_large_scale_alignment{pop_name}', 'instance'] == 'halo_mass':
                    beta = block[f'ia_large_scale_alignment{pop_name}', 'beta']
                    M_pivot = block[f'ia_large_scale_alignment{pop_name}', 'M_pivot']
                    c_align_profile = pk_lib.central_alignment_profile(mass, scale_factor, growth_factor,
                                                                      f_cen, C1, beta, M_pivot, mass_avg)
                else:
                    c_align_profile = pk_lib.central_alignment_profile(mass, scale_factor, growth_factor, f_cen, C1)
                # TODO: does this need the A_term?
                I_c_align_term = pk_lib.Ig_align_term(mass, c_align_profile, b_dm, dn_dlnm, mean_density0, A_term)
                I_s_align_term = pk_lib.Ig_align_term(mass, s_align_profile, b_dm, dn_dlnm, mean_density0, A_term)

                if bnl:
                    if p_mI:
                        I_NL_ia_gm = pk_lib.I_NL(mass, mass, c_align_profile + s_align_profile, matter_profile, b_dm, b_dm,
                                                 dn_dlnm, dn_dlnm, A_term, mean_density0, beta_interp, integrand_12, integrand_21, integrand_22)

                    if p_II:
                        I_NL_ia_cc = pk_lib.I_NL(mass, mass, c_align_profile, c_align_profile, b_dm, b_dm,
                                                 dn_dlnm, dn_dlnm, A_term, mean_density0, beta_interp, integrand_12, integrand_21, integrand_22)
                        I_NL_ia_cs = pk_lib.I_NL(mass, mass, c_align_profile, s_align_profile, b_dm, b_dm,
                                                 dn_dlnm, dn_dlnm, A_term, mean_density0, beta_interp, integrand_12, integrand_21, integrand_22)
                        I_NL_ia_ss = pk_lib.I_NL(mass, mass, s_align_profile, s_align_profile, b_dm, b_dm,
                                                 dn_dlnm, dn_dlnm, A_term, mean_density0, beta_interp, integrand_12, integrand_21, integrand_22)

                    if p_gI:
                        I_NL_ia_gc = pk_lib.I_NL(mass, mass, c_align_profile, profile_c, b_dm, b_dm,
                                                 dn_dlnm, dn_dlnm, A_term, mean_density0, beta_interp, integrand_12, integrand_21, integrand_22)
                        I_NL_ia_gs = pk_lib.I_NL(mass, mass, s_align_profile, profile_c, b_dm, b_dm,
                                                 dn_dlnm, dn_dlnm, A_term, mean_density0, beta_interp, integrand_12, integrand_21, integrand_22)

            if p_gg:
                # First check if the Poisson distribution for the satellites is disturbed.
                poisson_par = {
                    'poisson_type': poisson_type,
                    'poisson': get_string_or_none(block, 'pk_parameters', 'poisson', default=None),
                    'M_0': get_string_or_none(block, 'pk_parameters', 'M_0', default=None),
                    'slope': get_string_or_none(block, 'pk_parameters', 'slope', default=None)
                }

                if bnl:  # If bnl is True use the beyond linear halo bias formalism
                    pk_gg_1h, pk_gg_2h, pk_gg, bg_linear = pk_lib.compute_p_gg_bnl(k_vec, plin,
                                                                                  mass, dn_dlnm, profile_c, profile_s, I_c, I_s, I_NL_gg,
                                                                                  mass_avg, poisson_par, one_halo_ktrunc)
                else:  # If bnl is not true then just use linear halo bias
                    """
                    For Poisson distributed satellites, we have
                    < N_sat (N_sat-1) > = <N_sat>^2 = lambda^2 = <N_sat^2> - <N_sat>^2
                    We define the Poisson parameters as:
                    Poisson = <N_sat(N_sat-1)>/<N_sat>^2
                    A super-Poissonian distribution is a probability distribution
                    that has a larger variance than a Poisson distribution with the same mean.
                    Conversely, a sub-Poissonian distribution has a smaller variance.
                    Why do we assume Poisson distribution for satellites?
                    Poisson distribution assumes:
                    1- The satellites are independent. The existence of one satellite does not
                       affect another satellite.
                    2- Two satellites cannot exist in the same location.
                    3- The average number of satellites per halo is independent of any occurrence?
                    """
                    pk_gg_1h, pk_gg_2h, pk_gg, bg_linear = pk_lib.compute_p_gg(k_vec, plin,
                                                                               mass, dn_dlnm, profile_c, profile_s, I_c, I_s, mass_avg,
                                                                               poisson_par, one_halo_ktrunc, two_halo_ktrunc)
                block.put_grid(f'galaxy_linear_bias{suffix}', 'z', z_vec, 'k_h', k_vec, 'bg_linear', bg_linear)
                if response:
                    block.put_grid(f'galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_gg_1h / pk_mm * pk_mm_in)
                    block.put_grid(f'galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_gg_2h / pk_mm * pk_mm_in)
                    block.put_grid(f'galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_gg / pk_mm * pk_mm_in)
                else:
                    block.put_grid(f'galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_gg_1h)
                    block.put_grid(f'galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_gg_2h)
                    block.put_grid(f'galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_gg)

            if p_gm:
                if bnl:  # If bnl is True use the beyond linear halo bias formalism
                    pk_1h, pk_2h, pk_gm, bgm_linear = pk_lib.compute_p_gm_bnl(k_vec, plin,
                                                                             mass, dn_dlnm, profile_c, profile_s, matter_profile_1h, I_c, I_s, I_m,
                                                                             I_NL_gm, one_halo_ktrunc)
                else:  # If bnl is not true then just use linear halo bias
                    pk_1h, pk_2h, pk_gm, bgm_linear = pk_lib.compute_p_gm(k_vec, plin,
                                                                         mass, dn_dlnm, profile_c, profile_s, matter_profile_1h, I_c, I_s, I_m,
                                                                         one_halo_ktrunc, two_halo_ktrunc)
                block.put_grid(f'galaxy_matter_linear_bias{suffix}', 'z', z_vec, 'k_h', k_vec, 'bgm_linear', bgm_linear)
                if response:
                    block.put_grid(f'matter_galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_1h / pk_mm * pk_mm_in)
                    block.put_grid(f'matter_galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_2h / pk_mm * pk_mm_in)
                    block.put_grid(f'matter_galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_gm / pk_mm * pk_mm_in)
                else:
                    block.put_grid(f'matter_galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_1h)
                    block.put_grid(f'matter_galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_2h)
                    block.put_grid(f'matter_galaxy_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_gm)

            if fortuna:
                # Only used in Fortuna et al. 2021 implementation of IA power spectra
                # Computes the effective power spectrum, mixing the linear and nonlinear ones:
                # Default in Fortuna et al. 2021 is the non-linear power spectrum, so t_eff defaults to 0
                #
                # (1.-t_eff)*pnl + t_eff*plin
                #
                # Load nonlinear power spectrum
                k_nl, p_nl = pk_lib.get_nonlinear_power_spectrum(block, z_vec)
                pnl = pk_lib.log_linear_interpolation_k(p_nl, k_nl, k_vec)
                t_eff = block.get_double('pk_parameters', 'linear_fraction_fortuna', default=0.0)
                pk_eff = (1.0 - t_eff) * pnl + t_eff * plin

            if p_II:
                if fortuna:
                    pk_II_1h, pk_II_2h, pk_II = pk_lib.compute_p_II_fortuna(k_vec, pk_eff,
                                                                            mass, dn_dlnm, s_align_profile, alignment_amplitude_2h_II,
                                                                            f_cen, one_halo_ktrunc_ia, two_halo_ktrunc_ia)
                else:
                    if bnl:
                        pk_II_1h, pk_II_2h, pk_II = pk_lib.compute_p_II_bnl(k_vec, plin,
                                                                            mass, dn_dlnm, s_align_profile,
                                                                            I_c_align_term, I_s_align_term,
                                                                            I_NL_ia_cc, I_NL_ia_cs, I_NL_ia_ss,
                                                                            one_halo_ktrunc_ia)
                    else:
                        pk_II_1h, pk_II_2h, pk_II = pk_lib.compute_p_II(k_vec, plin,
                                                                        mass, dn_dlnm, s_align_profile,
                                                                        I_c_align_term, I_s_align_term,
                                                                        one_halo_ktrunc_ia, two_halo_ktrunc_ia)
                if response:
                    block.put_grid(f'intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_II_1h / pk_mm * pk_mm_in)
                    block.put_grid(f'intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_II_2h / pk_mm * pk_mm_in)
                    block.put_grid(f'intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_II / pk_mm * pk_mm_in)
                else:
                    block.put_grid(f'intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_II_1h)
                    block.put_grid(f'intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_II_2h)
                    block.put_grid(f'intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_II)

            if p_gI:
                if fortuna:
                    pk_gI_1h, pk_gI_2h, pk_gI = pk_lib.compute_p_gI_fortuna(k_vec, pk_eff,
                                                                            mass, dn_dlnm, profile_c, s_align_profile, I_c, alignment_amplitude_2h, one_halo_ktrunc_ia, two_halo_ktrunc_ia)
                else:
                    if bnl:
                        pk_gI_1h, pk_gI_2h, pk_gI = pk_lib.compute_p_gI_bnl(k_vec, plin,
                                                                            mass, dn_dlnm, profile_c, s_align_profile, I_c, I_c_align_term, I_s_align_term,
                                                                            I_NL_ia_gc, I_NL_ia_gs, one_halo_ktrunc_ia)
                    else:
                        pk_gI_1h, pk_gI_2h, pk_gI = pk_lib.compute_p_gI(k_vec, plin,
                                                                        mass, dn_dlnm, profile_c, s_align_profile, I_c,
                                                                        I_c_align_term, I_s_align_term, one_halo_ktrunc_ia, two_halo_ktrunc_ia)
                if response:
                    block.put_grid(f'galaxy_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_gI_1h / pk_mm * pk_mm_in)
                    block.put_grid(f'galaxy_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_gI_2h / pk_mm * pk_mm_in)
                    block.put_grid(f'galaxy_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_gI / pk_mm * pk_mm_in)
                else:
                    block.put_grid(f'galaxy_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_gI_1h)
                    block.put_grid(f'galaxy_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_gI_2h)
                    block.put_grid(f'galaxy_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_gI)

            if p_mI:
                if fortuna:
                    pk_mI_1h, pk_mI_2h, pk_mI = pk_lib.compute_p_mI_fortuna(k_vec, pk_eff,
                                                                            mass, dn_dlnm, matter_profile_1h, s_align_profile, alignment_amplitude_2h,
                                                                            f_cen, one_halo_ktrunc_ia, two_halo_ktrunc_ia)
                else:
                    if bnl:
                        pk_mI_1h, pk_mI_2h, pk_mI = pk_lib.compute_p_mI_bnl(k_vec, plin,
                                                                            mass, dn_dlnm, matter_profile_1h, s_align_profile, I_m, I_c_align_term, I_s_align_term,
                                                                            I_NL_ia_gm, one_halo_ktrunc_ia)
                    else:
                        pk_mI_1h, pk_mI_2h, pk_mI = pk_lib.compute_p_mI(k_vec, plin,
                                                                        mass, dn_dlnm, matter_profile_1h, s_align_profile, I_m, I_c_align_term, I_s_align_term,
                                                                        one_halo_ktrunc_ia, two_halo_ktrunc_ia)
                if response:
                    block.put_grid(f'matter_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_mI_1h / pk_mm * pk_mm_in)
                    block.put_grid(f'matter_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_mI_2h / pk_mm * pk_mm_in)
                    block.put_grid(f'matter_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_mI / pk_mm * pk_mm_in)
                else:
                    block.put_grid(f'matter_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_mI_1h)
                    block.put_grid(f'matter_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_mI_2h)
                    block.put_grid(f'matter_intrinsic_power{suffix}', 'z', z_vec, 'k_h', k_vec, 'p_k', pk_mI)

    return 0

def cleanup(config):
    # Usually python modules do not need to do anything here.
    # We just leave it in out of pedantic completeness.
    pass



"""

`pyfftlog` -- Python version of FFTLog
======================================

This is a Python version of the FFTLog Fortran code by Andrew Hamilton:

Hamilton, A. J. S., 2000, Uncorrelated modes of the non-linear power spectrum:
Monthly Notices of the Royal Astronomical Society, 312, pages 257-284; DOI:
10.1046/j.1365-8711.2000.03071.x; Website of FFTLog:
http://casa.colorado.edu/~ajsh/FFTLog.

The function `scipy.special.loggamma` replaces the file `cdgamma.f` in the
original code, and the functions `rfft` and `irfft` from `scipy.fftpack`
replace the files `drffti.f`, `drfftf.f`, and `drfftb.f` in the original code.

The original documentation has just been adjusted where necessary, and put into
a more pythonic format (e.g. using 'Parameters' and 'Returns' in the
documentation').

***** From the original documentation *****

===============================================================================
 THE FFTLog CODE
===============================================================================

FFTLog
    computes the discrete Fast Fourier Transform or Fast Hankel Transform (of
    arbitrary real index) of a periodic logarithmic sequence.

Version of 5 July 1999.

For more information about FFTLog, see http://casa.colorado.edu/~ajsh/FFTLog.

Andrew J S Hamilton March 1999, email: Andrew.Hamilton@Colorado.EDU

Refs: Talman J. D., 1978, J. Comp. Phys., 29, 35; Hamilton A. J. S., 2000,
      MNRAS, 312, 257 (http://xxx.lanl.gov/abs/astro-ph/9905191)

FFTLog computes a discrete version of the Hankel Transform (= Fourier-Bessel
Transform) with a power law bias (k r)^q

            infinity
            /           q
    ã(k) = | a(r) (k r)  J  (k r) k dr
            /              mu
            0

            infinity
            /           -q
    a(r) = | ã(k) (k r)   J  (k r) r dk
            /               mu
            0

where J_mu is the Bessel function of order mu.  The index mu may be any real
number, positive or negative.

The input array a_j is a periodic sequence of length n, uniformly
logarithmically spaced with spacing dlnr
    a_j = a(r_j)   at   r_j = r_c exp[(j-j_c) dlnr]
centred about the point r_c.  The central index j_c = (n+1)/2 is 1/2 integral
if n is even.  Similarly, the output array ã_j is a periodic sequence of length
n, also uniformly logarithmically spaced with spacing dlnr
    ã_j = ã(k_j)   at   k_j = k_c exp[(j-j_c) dlnr]
centred about the point k_c.

The centre points r_c and k_c of the periodic intervals may be chosen
arbitrarily; but it would be normal to choose the product
    kr = k_c r_c = k_j r_(n+1-j) = k_(n+1-j) r_j
to be about 1 (or 2, or pi, to taste).

The FFTLog algorithm is (see Hamilton 2000):
1. FFT the input array a_j to obtain the Fourier coefficients c_m ;
2. Multiply c_m by
        u_m = (kr)^[- i 2 m pi/(n dlnr)] U_mu[q + i 2 m pi/(n dlnr)]
    where
        U_mu(x) = 2^x Gamma[(mu+1+x)/2] / Gamma[(mu+1-x)/2]
    to obtain c_m u_m ;
3. FFT c_m u_m back to obtain the discrete Hankel transform ã_j .

----------------------------------------------------------------------
The Fourier sine and cosine transforms

                    infinity
                        /
    Ã(k) = sqrt(2/pi) | A(r) sin(k r) dr
                        /
                    0

                    infinity
                        /
    Ã(k) = sqrt(2/pi) | A(r) cos(k r) dr
                        /
                    0

may be regarded as special cases of the Hankel transform with mu = 1/2 and -1/2
since

    sqrt(2/pi) sin(x) = sqrt(x) J   (x)
                                1/2

    sqrt(2/pi) cos(x) = sqrt(x) J    (x)
                                -1/2

The Fourier transforms may be done by making the substitutions
                q-(1/2)                      -q-(1/2)
    A(r) = a(r) r          and   Ã(k) = ã(k) k

and Hankel transforming a(r) with a power law bias (k r)^q

            infinity
            /           q
    ã(k) = | a(r) (k r)  J    (k r) k dr
            /              ±1/2
            0

Different choices of power law bias q lead to different discrete Fourier
transforms of A(r), because the assumption of periodicity of a(r) = A(r)
r^[-q+(1/2)] is different for different q.

If A(r) is a power law, A(r) proportional to r^[q-(1/2)], then applying a bias
q yields a discrete Fourier transform Ã(k) that is exactly equal to the
continuous Fourier transform, because then a(r) is a constant, which is a
periodic function.

----------------------------------------------------------------------
The Hankel transform

            infinity
            /
    Ã(k) = | A(r) J  (k r) k dr
            /       mu
            0

may be done by making the substitutions
                q                      -q
    A(r) = a(r) r    and   Ã(k) = ã(k) k

and Hankel transforming a(r) with a power law bias (k r)^q

            infinity
            /           q
    ã(k) = | a(r) (k r)  J  (k r) k dr
            /              mu
            0

Different choices of power law bias q lead to different discrete Hankel
transforms of A(r), because the assumption of periodicity of a(r) = A(r) r^-q
is different for different q.

If A(r) is a power law, A(r) proportional to r^q, then applying a bias q yields
a discrete Hankel transform Ã(k) that is exactly equal to the continuous Hankel
transform, because then a(r) is a constant, which is a periodic function.

----------------------------------------------------------------------
------------------------
There are five routines:
------------------------
Comments in the subroutines contain further details.

(1) subroutine fhti(n,mu,q,dlnr,kr,kropt,wsave,ok)
    is an initialization routine.

(2) subroutine fftl(n,a,rk,dir,wsave)
    computes the discrete Fourier sine or cosine transform of a logarithmically
    spaced periodic sequence. This is a driver routine that calls fhtq.

(3) subroutine fht(n,a,dir,wsave)
    computes the discrete Hankel transform of a logarithmically spaced periodic
    sequence.  This is a driver routine that calls fhtq.

(4) subroutine fhtq(n,a,dir,wsave)
    computes the biased discrete Hankel transform of a logarithmically spaced
    periodic sequence.  This is the basic FFTLog routine.

(5) real*8 function krgood(mu,q,dlnr,kr)
    takes an input kr and returns the nearest low-ringing kr.  This is an
    optional routine called by fhti.

"""

import numpy as np

# from scipy.fftpack._fftpack import drfft
from scipy.fftpack import irfft, rfft
from scipy.special import loggamma

# from scipy.fftpack.basic import _raw_fft


def fhti(n, mu, dlnr, q=0, kr=1, kropt=0):
    """Initialize the working array xsave used by fftl, fht, and fhtq.

    fhti initializes the working array xsave used by fftl, fht, and fhtq.  fhti
    need be called once, whereafter fftl, fht, or fhtq may be called many
    times, as long as n, mu, q, dlnr, and kr remain unchanged. fhti should be
    called each time n, mu, q, dlnr, or kr is changed. The work array xsave
    should not be changed between calls to fftl, fht, or fhtq.

    Parameters
    ----------
    n : int
        Number of points in the array to be transformed; n may be any positive
        integer, but the FFT routines run fastest if n is a product of small
        primes 2, 3, 5.

    mu : float
        Index of J_mu in Hankel transform; mu may be any real number, positive
        or negative.

    dlnr : float
        Separation between natural log of points; dlnr may be positive or
        negative.

    q : float, optional
        Exponent of power law bias; q may be any real number, positive or
        negative.  If in doubt, use q = 0, for which case the Hankel transform
        is orthogonal, i.e. self-inverse, provided also that, for n even, kr is
        low-ringing.  Non-zero q may yield better approximations to the
        continuous Hankel transform for some functions.
        Defaults to 0 (unbiased).

    kr : float, optional
        k_c r_c where c is central point of array
        = k_j r_(n+1-j) = k_(n+1-j) r_j .
        Normally one would choose kr to be about 1 (default) (or 2, or pi, to
        taste).

    kropt : int, optional; {0, 1, 2, 3}
        - 0 to use input kr as is (default);
        - 1 to change kr to nearest low-ringing kr, quietly;
        - 2 to change kr to nearest low-ringing kr, verbosely;
        - 3 for option to change kr interactively.

    Returns
    -------
    kr : float, optional
        kr, adjusted depending on kropt.

    xsave : array
        Working array used by fftl, fht, and fhtq. Dimension:
        - for q = 0 (unbiased transform): n+3
        - for q != 0 (biased transform): 1.5*n+4
        If odd, last element is not needed.

    """

    # adjust kr
    if kropt == 0:    # keep kr as is
        pass
    elif kropt == 1:  # change kr to low-ringing kr quietly
        kr = krgood(mu, q, dlnr, kr)
    elif kropt == 2:  # change kr to low-ringing kr verbosely
        d = krgood(mu, q, dlnr, kr)
        if abs(kr / d - 1) >= 1e-15:
            kr = d
            print((" kr changed to ", kr))
    else:             # option to change kr to low-ringing kr interactively
        d = krgood(mu, q, dlnr, kr)
        if abs(kr / d - 1.0) >= 1e-15:
            print((" change kr = ", kr))
            print((" to low-ringing kr = ", d))
            go = eval(input("? [CR, y=yes, n=no, x=exit]: "))
            if go.lower() in ['', 'y']:
                kr = d
                print((" kr changed to ", kr))
            elif go.lower() == 'n':
                print((" kr left unchanged at ", kr))
            else:
                print("exit")
                return False

    # return if n is <= 0
    if n <= 0:
        return kr

    # The normal FFT is not initialized here as in the original FFTLog code, as
    # the `scipy.fftpack`-FFT routines `rfft` and `irfft` do that internally.
    # Therefore xsave in `pyfftlog` is 2*n+15 elements shorter, and named
    # xsave to not confuse it with xsave from the FFT.

    if q == 0:  # unbiased case (q = 0)
        ln2kr = np.log(2.0 / kr)
        xp = (mu + 1) / 2.0
        d = np.pi / (n * dlnr)

        m = np.arange(1, (n + 1) / 2)
        y = m * d  # y = m*pi/(n*dlnr)
        zp = loggamma(xp + 1j * y)
        # Argument of kr^(-2 i y) U_mu(2 i y)
        arg = 2.0 * (ln2kr * y + zp.imag)

        # Arange xsave: [q, dlnr, kr, cos, sin]
        xsave = np.empty(2 * arg.size + 3)
        xsave[0] = q
        xsave[1] = dlnr
        xsave[2] = kr
        xsave[3::2] = np.cos(arg)
        xsave[4::2] = np.sin(arg)

        # Altogether 3 + 2*(n/2) elements used for q = 0, which is n+3 for even
        # n, n+2 for odd n.

    else:       # biased case (q != 0)
        ln2 = np.log(2.0)
        ln2kr = np.log(2.0 / kr)
        xp = (mu + 1 + q) / 2.0
        xm = (mu + 1 - q) / 2.0

        # first element of rest of xsave
        y = 0

        # case where xp or xm is a negative integer
        xpnegi = np.round(xp) == xp and xp <= 0
        xmnegi = np.round(xm) == xm and xm <= 0
        if xpnegi or xmnegi:

            # case where xp and xm are both negative integers
            # U_mu(q) = 2^q Gamma[xp]/Gamma[xm] is finite in this case
            if xpnegi and xmnegi:
                # Amplitude and Argument of U_mu(q)
                amp = np.exp(ln2 * q)
                if xp > xm:
                    m = np.arange(1,  np.round(xp - xm) + 1)
                    amp *= xm + m - 1
                elif xp < xm:
                    m = np.arange(1,  np.round(xm - xp) + 1)
                    amp /= xp + m - 1
                arg = np.round(xp + xm) * np.pi

            else:  # one of xp or xm is a negative integer
                # Transformation is singular if xp is -ve integer, and inverse
                # transformation is singular if xm is -ve integer, but
                # transformation may be well-defined if sum_j a_j = 0, as may
                # well occur in physical cases.  Policy is to drop the
                # potentially infinite constant in the transform.

                if xpnegi:
                    print(('fhti: (mu+1+q)/2 =', np.round(xp), 'is -ve integer',
                          ', yields singular transform:\ntransform will omit',
                          'additive constant that is generically infinite,',
                          '\nbut that may be finite or zero if the sum of the',
                          'elements of the input array a_j is zero.'))
                else:
                    print(('fhti: (mu+1-q)/2 =', np.round(xm), 'is -ve integer',
                          ', yields singular inverse transform:\n inverse',
                          'transform will omit additive constant that is',
                          'generically infinite,\nbut that may be finite or',
                          'zero if the sum of the elements of the input array',
                          'a_j is zero.'))
                amp = 0
                arg = 0

        else:  # neither xp nor xm is a negative integer
            zp = loggamma(xp + 1j * y)
            zm = loggamma(xm + 1j * y)

            # Amplitude and Argument of U_mu(q)
            amp = np.exp(ln2 * q + zp.real - zm.real)
            # note +Im(zm) to get conjugate value below real axis
            arg = zp.imag + zm.imag

        # first element: cos(arg) = ±1, sin(arg) = 0
        xsave1 = amp * np.cos(arg)

        # remaining elements of xsave
        d = np.pi / (n * dlnr)
        m = np.arange(1, (n + 1) / 2)
        y = m * d  # y = m pi/(n dlnr)
        zp = loggamma(xp + 1j * y)
        zm = loggamma(xm + 1j * y)
        # Amplitude and Argument of kr^(-2 i y) U_mu(q + 2 i y)
        amp = np.exp(ln2 * q + zp.real - zm.real)
        arg = 2 * ln2kr * y + zp.imag + zm.imag

        # Arrange xsave: [q, dlnr, kr, xsave1, cos, sin]
        xsave = np.empty(3 * arg.size + 4)
        xsave[0] = q
        xsave[1] = dlnr
        xsave[2] = kr
        xsave[3] = xsave1
        xsave[4::3] = amp
        xsave[5::3] = np.cos(arg)
        xsave[6::3] = np.sin(arg)

        # Altogether 3 + 3*(n/2)+1 elements used for q != 0, which is (3*n)/2+4
        # for even n, (3*n)/2+3 for odd n.  For even n, the very last element
        # of xsave [i.e. xsave(3*m+1)=sin(arg) for m=n/2] is not used within
        # FFTLog; if a low-ringing kr is used, this element should be zero.
        # The last element is computed in case somebody wants it.

    return kr, xsave


def fftl(a, xsave, rk=1, tdir=1):
    """Logarithmic fast Fourier transform FFTLog.

    This is a driver routine that calls fhtq.

    fftl computes a discrete version of the Fourier sine (if mu = 1/2) or
    cosine (if mu = -1/2) transform

                        infinity
                         /
       Ã(k) = sqrt(2/pi) | A(r) sin(k r) dr
                         /
                        0

                        infinity
                         /
       Ã(k) = sqrt(2/pi) | A(r) cos(k r) dr
                         /
                        0

    by making the substitutions
                    q-(1/2)                      -q-(1/2)
       A(r) = a(r) r          and   Ã(k) = ã(k) k

    and applying a biased Hankel transform to a(r).

    The steps are:
    1. a(r) = A(r) r^[-dir*(q-.5)]
    2. call fhtq to transform a(r) -> ã(k)
    3. Ã(k) = ã(k) k^[-dir*(q+.5)]

    fhti must be called before the first call to fftl, with mu = 1/2 for a sine
    transform, or mu = -1/2 for a cosine transform.

    A call to fftl with dir=1 followed by a call to fftl with dir=-1 (and rk
    unchanged), or vice versa, leaves the array a unchanged.

    Parameters
    ----------
    a : array
        Array A(r) to transform: a(j) is A(r_j) at r_j = r_c exp[(j-jc) dlnr],
        where jc = (n+1)/2 = central index of array.

    xsave : array
        Working array set up by fhti.

    rk : float, optional
        r_c/k_c = r_j/k_j (a constant, the same constant for any j); rk is not
        (necessarily) the same quantity as kr.  rk is used only to multiply the
        output array by sqrt(rk)^dir, so if you want to do the normalization
        later, or you don't care about the normalization, you can set rk = 1.
        Defaults to 1.

    tdir : int, optional; {1, -1}
        -  1 for forward transform (default),
        - -1 for backward transform.
        A backward transform (dir = -1) is the same as a forward transform with
        q -> -q and rk -> 1/rk, for any kr if n is odd, for low-ringing kr if n
        is even.

    Returns
    -------
    a : array
        Transformed array Ã(k): a(j) is Ã(k_j) at k_j = k_c exp[(j-jc) dlnr].

    """
    fct = a.copy()
    q = xsave[0]
    dlnr = xsave[1]
    kr = xsave[2]

    # centre point of array
    jc = np.array((fct.size + 1) / 2.0)
    j = np.arange(fct.size) + 1

    # a(r) = A(r) (r/rc)^[-dir*(q-.5)]
    fct *= np.exp(-tdir * (q - 0.5) * (j - jc) * dlnr)

    # transform a(r) -> ã(k)
    fct = fhtq(fct, xsave, tdir)

    # Ã(k) = ã(k) k^[-dir*(q+.5)] rc^[-dir*(q-.5)]
    #      = ã(k) (k/kc)^[-dir*(q+.5)] (kc rc)^(-dir*q) (rc/kc)^(dir*.5)
    lnkr = np.log(kr)
    lnrk = np.log(rk)
    fct *= np.exp(-tdir * ((q + 0.5) * (j - jc)
                           * dlnr + q * lnkr - lnrk / 2.0))

    return fct


def fht(a, xsave, tdir=1):
    """Fast Hankel transform FHT.

    This is a driver routine that calls fhtq.

    fht computes a discrete version of the Hankel transform

             infinity
              /
       Ã(k) = | A(r) J  (k r) k dr
              /       mu
             0

    by making the substitutions
                    q                      -q
       A(r) = a(r) r    and   Ã(k) = ã(k) k

    and applying a biased Hankel transform to a(r).

    The steps are:
    1. a(r) = A(r) r^(-dir*q)
    2. call fhtq to transform a(r) -> ã(k)
    3. Ã(k) = ã(k) k^(-dir*q)

    fhti must be called before the first call to fht.

    A call to fht with dir=1 followed by a call to fht with dir=-1, or vice
    versa, leaves the array a unchanged.



    Parameters
    ----------
    a : array
        Array A(r) to transform: a(j) is A(r_j) at r_j = r_c exp[(j-jc) dlnr],
        where jc = (n+1)/2 = central index of array.

    xsave : array
        Working array set up by fhti.

    tdir : int, optional; {1, -1}
        -  1 for forward transform (default),
        - -1 for backward transform.
        A backward transform (dir = -1) is the same as a forward transform with
        q -> -q, for any kr if n is odd, for low-ringing kr if n is even.

    Returns
    -------
    a : array
        Transformed array Ã(k): a(j) is Ã(k_j) at k_j = k_c exp[(j-jc) dlnr].

    """
    fct = a.copy()
    q = xsave[0]
    dlnr = xsave[1]
    kr = xsave[2]

    # a(r) = A(r) (r/rc)^(-dir*q)
    if q != 0:
        #  centre point of array
        jc = np.array((fct.size + 1) / 2.0)
        j = np.arange(fct.size) + 1
        fct *= np.exp(-tdir * q * (j - jc) * dlnr)

    # transform a(r) -> ã(k)
    fct = fhtq(fct, xsave, tdir)

    # Ã(k) = ã(k) (k rc)^(-dir*q)
    #      = ã(k) (k/kc)^(-dir*q) (kc rc)^(-dir*q)
    if q != 0:
        lnkr = np.log(kr)
        fct *= np.exp(-tdir * q * ((j - jc) * dlnr + lnkr))

    return fct


def fhtq(a, xsave, tdir=1):
    """Kernel routine of FFTLog.

    This is the basic FFTLog routine.

    fhtq computes a discrete version of the biased Hankel transform

             infinity
              /           q
       ã(k) = | a(r) (k r)  J  (k r) k dr
              /              mu
             0

    fhti must be called before the first call to fhtq.

    A call to fhtq with dir=1 followed by a call to fhtq with dir=-1, or vice
    versa, leaves the array a unchanged.

    Parameters
    ----------
    a : array
        Periodic array a(r) to transform: a(j) is a(r_j) at r_j = r_c
        exp[(j-jc) dlnr] where jc = (n+1)/2 = central index of array.

    xsave : array
        Working array set up by fhti.

    tdir : int, optional; {1, -1}
        -  1 for forward transform (default),
        - -1 for backward transform.
        A backward transform (dir = -1) is the same as a forward transform with
        q -> -q, for any kr if n is odd, for low-ringing kr if n is even.

    Returns
    -------
    a : array
        Transformed periodic array ã(k): a(j) is ã(k_j) at k_j = k_c exp[(j-jc)
        dlnr].

    """
    fct = a.copy()
    q = xsave[0]
    n = fct.size

    # normal FFT
    fct = rfft(fct)
    # _raw_fft(fct, n, -1, 1, 1, _fftpack.drfft)
    # breakpoint()
    # fct = drfft(fct, n, 1, 0)

    m = np.arange(1, n / 2, dtype=int)  # index variable
    if q == 0:  # unbiased (q = 0) transform
        # multiply by (kr)^[- i 2 m pi/(n dlnr)] U_mu[i 2 m pi/(n dlnr)]
        ar = fct[2 * m - 1]
        ai = fct[2 * m]
        fct[2 * m - 1] = ar * xsave[2 * m + 1] - ai * xsave[2 * m + 2]
        fct[2 * m] = ar * xsave[2 * m + 2] + ai * xsave[2 * m + 1]
        # problem(2*m)atical last element, for even n
        if np.mod(n, 2) == 0:
            ar = xsave[-2]
            if (tdir == 1):  # forward transform: multiply by real part
                # Why? See http://casa.colorado.edu/~ajsh/FFTLog/index.html#ure
                fct[-1] *= ar
            elif (tdir == -1):  # backward transform: divide by real part
                # Real part ar can be zero for maximally bad choice of kr.
                # This is unlikely to happen by chance, but if it does, policy
                # is to let it happen.  For low-ringing kr, imaginary part ai
                # is zero by construction, and real part ar is guaranteed
                # nonzero.
                fct[-1] /= ar

    else:  # biased (q != 0) transform
        # multiply by (kr)^[- i 2 m pi/(n dlnr)] U_mu[q + i 2 m pi/(n dlnr)]
        # phase
        ar = fct[2 * m - 1]
        ai = fct[2 * m]
        fct[2 * m - 1] = ar * xsave[3 * m + 2] - ai * xsave[3 * m + 3]
        fct[2 * m] = ar * xsave[3 * m + 3] + ai * xsave[3 * m + 2]

        if tdir == 1:  # forward transform: multiply by amplitude
            fct[0] *= xsave[3]
            fct[2 * m - 1] *= xsave[3 * m + 1]
            fct[2 * m] *= xsave[3 * m + 1]

        elif tdir == -1:  # backward transform: divide by amplitude
            # amplitude of m=0 element
            ar = xsave[3]
            if ar == 0:
                # Amplitude of m=0 element can be zero for some mu, q
                # combinations (singular inverse); policy is to drop
                # potentially infinite constant.
                fct[0] = 0
            else:
                fct[0] /= ar

            # remaining amplitudes should never be zero
            fct[2 * m - 1] /= xsave[3 * m + 1]
            fct[2 * m] /= xsave[3 * m + 1]

        # problematical last element, for even n
        if np.mod(n, 2) == 0:
            m = int(n / 2)
            ar = xsave[3 * m + 2] * xsave[3 * m + 1]
            if tdir == 1:  # forward transform: multiply by real part
                fct[-1] *= ar
            elif (tdir == -1):  # backward transform: divide by real part
                # Real part ar can be zero for maximally bad choice of kr.
                # This is unlikely to happen by chance, but if it does, policy
                # is to let it happen.  For low-ringing kr, imaginary part ai
                # is zero by construction, and real part ar is guaranteed
                # nonzero.
                fct[-1] /= ar

    # normal FFT back
    fct = irfft(fct)
    # _raw_fft(fct, n, -1, -1, 1, _fftpack.drfft)
    # fct = drfft(fct, n, -1, 1)

    # reverse the array and at the same time undo the FFTs' multiplication by n
    # => Just reverse the array, the rest is already done in drfft.
    fct = fct[::-1]

    return fct


def krgood(mu, q, dlnr, kr):
    """Return optimal kr

    Use of this routine is optional.

    Choosing kr so that
        (kr)^(- i pi/dlnr) U_mu(q + i pi/dlnr)
    is real may reduce ringing of the discrete Hankel transform, because it
    makes the transition of this function across the period boundary smoother.

    Parameters
    ----------
    mu : float
        index of J_mu in Hankel transform; mu may be any real number, positive
        or negative.

    q : float
        exponent of power law bias; q may be any real number, positive or
        negative.  If in doubt, use q = 0, for which case the Hankel transform
        is orthogonal, i.e. self-inverse, provided also that, for n even, kr is
        low-ringing.  Non-zero q may yield better approximations to the
        continuous Hankel transform for some functions.

    dlnr : float
        separation between natural log of points; dlnr may be positive or
        negative.

    kr : float, optional
        k_c r_c where c is central point of array
        = k_j r_(n+1-j) = k_(n+1-j) r_j .
        Normally one would choose kr to be about 1 (default) (or 2, or pi, to
        taste).

    Returns
    -------
    krgood : float
        low-ringing value of kr nearest to input kr.  ln(krgood) is always
        within dlnr/2 of ln(kr).

    """
    if dlnr == 0:
        return kr

    xp = (mu + 1.0 + q) / 2.0
    xm = (mu + 1.0 - q) / 2.0
    y = 1j * np.pi / (2.0 * dlnr)
    zp = loggamma(xp + y)
    zm = loggamma(xm + y)

    # low-ringing condition is that following should be integral
    arg = np.log(2.0 / kr) / dlnr + (zp.imag + zm.imag) / np.pi

    # return low-ringing kr
    return kr * np.exp((arg - np.round(arg)) * dlnr)
